#!/usr/bin/env python
# encoding: utf-8

import os
import re
import sys
import time
import json
import pprint
import hashlib
import argparse
import subprocess

from abc import ABCMeta, abstractmethod
from collections import namedtuple, Counter


CFG = namedtuple('Config', [
    # How often to run the program on the dataset.
    # The fs cache is flushed only the first time.
    'n_runs',

    # How many times to run each individual program.
    'n_sub_runs',

    # Output directory
    'output'
])(3, 2, 'bench-output-{stamp}'.format(
    stamp=time.strftime('%FT%T%z', time.localtime())
))


def flush_fs_caches():
    os.sync()
    try:
        with open('/proc/sys/vm/drop_caches', 'w') as handle:
            handle.write('3\n')
        print('-- Flushed fs cache.')
    except IOError:
        print('!! You need to be root to flush the fs cache.')
        sys.exit(-1)


def read_time_info():
    try:
        with open('/tmp/.cpu_usage', 'r') as handle:
            cpu_usage, peak_mem = handle.read().strip().split(' ', 1)
            return int(cpu_usage[:-1]), int(peak_mem)
    except IOError:
        return 0


class Program:
    __metaclass__ = ABCMeta

    version = ''
    binary_path = None
    website = None
    script = None

    def __init__(self):
        # Do the path building now.
        # Later on we call chdir() which might mess this up:
        self.build_path = os.path.join(
            os.path.realpath(os.path.dirname(__file__)),
            'build_scripts',
            self.script or ''
        )

    def get_install(self):
        if self.script:
            with open(self.build_path, 'r') as fd:
                return fd.read()

    ####################
    # ABSTRACT METHODS #
    ####################

    @abstractmethod
    def get_binary(self):
        pass

    @abstractmethod
    def compute_version(self):
        pass

    def get_options(self, paths):
        return ''

    def get_benchid(self):
        return self.get_name()

    ####################
    # COMMON FUNCTIONS #
    ####################

    def get_name(self):
        """Get a printable name of the name (e.g. "rmlint")
        """
        return os.path.basename(self.get_binary())

    def get_temp_dir(self):
        temp_hash = '-'.join((
            self.get_name(),
            hashlib.md5(self.get_binary().encode('ascii')).hexdigest()[:6]
        ))
        temp_dir = os.path.join('/tmp', 'rmlint-bench', temp_hash)

        try:
            os.makedirs(temp_dir)
        except OSError:
            pass

        return temp_dir

    def get_website(self):
        if self.website:
            return self.website
        else:
            return 'https://www.google.de/search?q=' + self.get_name()

    def run(self, dataset):
        """Run the program on a given dataset
        """
        run_benchmarks = {}
        memory_usage = -1

        paths = dataset.get_paths()
        bin_cmd = self.binary_path + ' ' + self.get_options(paths)

        print('== Executing: {c}'.format(c=bin_cmd))

        time_cmd = '/bin/time --format "%P %M" \
            --output /tmp/.cpu_usage -- ' + bin_cmd

        try:
            for _ in range(CFG.n_sub_runs):
                flush_fs_caches()

                for run_idx in range(1, CFG.n_runs + 1):
                    print('.. Doing run #' + str(run_idx))
                    start_time = time.time()
                    data_dump = subprocess.check_output(
                        time_cmd,
                        shell=True,
                        stderr=subprocess.STDOUT
                    )

                    time_diff = time.time() - start_time

                    # Remember the time difference as result.
                    if run_idx not in run_benchmarks:
                        run_benchmarks[run_idx] = [0] * 5

                    cpu_usage, peak_mem = read_time_info()

                    # select the fastest run to record time and cpu stats:
                    if run_benchmarks[run_idx][0] > time_diff or run_benchmarks[run_idx][0] == 0:
                        run_benchmarks[run_idx][0] = time_diff
                        run_benchmarks[run_idx][1] = cpu_usage
                        run_benchmarks[run_idx][2] = peak_mem / 1024  # Megabyte

                    if data_dump:
                        stats = self.parse_statistics(data_dump)
                        if stats:
                            run_benchmarks[run_idx][3] = stats['dupes']
                            run_benchmarks[run_idx][4] = stats['sets']

        except subprocess.CalledProcessError as err:
            print('!! Execution of {n} failed: {err}'.format(
                n=self.binary_path, err=err
            ))

        avg_point = [0] * 5
        for idx in range(5):
            avg_point[idx] = sum([v[idx] for v in run_benchmarks.values()])
            avg_point[idx] /= CFG.n_runs

        print('== Took avg time of {t}s / {c}% cpu'.format(
            t=avg_point[0], c=avg_point[1]
        ))
        run_benchmarks['average'] = avg_point

        return run_benchmarks, memory_usage

    def guess_version(self):
        version = self.compute_version()
        if version is not None:
            self.version = str(version).strip()
            print('-- Guessed version: ', self.version)

    def install(self):
        """Install the program in /tmp; fetch source from internet."""
        temp_dir = self.get_temp_dir()
        temp_bin = os.path.join(temp_dir, self.get_binary())
        current_path = os.getcwd()

        os.chdir(temp_dir)
        self.binary_path = os.path.abspath(self.get_binary())

        if os.path.exists(temp_bin):
            print('-- Path exists: ' + temp_bin)
            print('-- Skipping install; Delete if you need an update.')
            self.guess_version()
            os.chdir(current_path)
            return

        sh_procedure = self.get_install()
        if not sh_procedure:
            return

        try:
            subprocess.check_output(
                sh_procedure, stderr=subprocess.STDOUT, shell=True
            )
            print('-- Installed {n} to {p}'.format(
                n=self.get_name(), p=temp_dir
            ))

            self.guess_version()
            return self.binary_path
        except subprocess.CalledProcessError as err:
            print('!! Failed to install {n}: {err.output}'.format(
                n=self.get_name(), err=err
            ))
            self.binary_path = None
            return None
        finally:
            os.chdir(current_path)


class Rmlint(Program):
    website = 'https://github.com/sahib/rmlint'
    script = 'rmlint-new.sh'

    def get_binary(self):
        return 'rmlint/rmlint'

    def get_options(self, paths):
        return '--hidden -o summary -o json:/tmp/rmlint.json -T df ' + ' '.join(paths)

    def compute_version(self):
        version_text = subprocess.check_output(
            self.get_binary() + ' --version', shell=True, stderr=subprocess.STDOUT
        )

        match = re.search('(\d.\d.\d) .*(rev [0-9a-f]{7})', str(version_text))
        if match is not None:
            return ' '.join(match.groups())

        return ""

    def parse_statistics(self, _):
        try:
            with open('/tmp/rmlint.json', 'r') as fd:
                stats = json.loads(fd.read())

            return {
                'sets': stats[-1]['duplicate_sets'],
                'dupes': stats[-1]['duplicates']
            }
        except OSError:
            pass


class RmlintSpot(Rmlint):
    website = 'https://github.com/SeeSpotRun/rmlint'
    script = 'rmlint-spot.sh'

    def get_binary(self):
        return 'rmlint/rmlint-spot'

    def get_benchid(self):
        return 'rmlint-spot'

class RmlintSpotParanoid(RmlintSpot):
    def get_options(self, paths):
        return '-p ' + RmlintSpot.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-spot-paranoid'


class Rmlint222(Rmlint):
    script = 'rmlint-v2.2.2.sh'

    def get_binary(self):
        return 'rmlint/rmlint-v2.2.2'

    def get_benchid(self):
        return 'rmlint-v2.2.2'


class Rmlint246(Rmlint):
    script = 'rmlint-v2.4.6.sh'

    def get_binary(self):
        return 'rmlint/rmlint-v2.4.6'

    def get_benchid(self):
        return 'rmlint-v2.4.6'



class Rmlint222Paranoid(Rmlint222):
    def get_options(self, paths):
        return '-p ' + Rmlint222.get_options(self, paths)

    def get_benchid(self):
        return Rmlint222.get_benchid(self) + '-paranoid'


class RmlintMaster(Rmlint):
    script = 'rmlint-master.sh'

    def get_binary(self):
        return 'rmlint/rmlint-master'

    def get_benchid(self):
        return 'rmlint-master'


class RmlintSpooky(Rmlint):
    def get_options(self, paths):
        return '-a spooky ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-spooky'


class RmlintXXHash(Rmlint):
    def get_options(self, paths):
        return '-a xxhash ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-xxhash'


class RmlintCity(Rmlint):
    def get_options(self, paths):
        return '-a city ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-city'


class RmlintMD5(Rmlint):
    def get_options(self, paths):
        return '-a md5 ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-md5'


class RmlintMurmur(Rmlint):
    def get_options(self, paths):
        return '-a murmur ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-murmur'


class RmlintParanoid(Rmlint):

    def get_options(self, paths):
        return '-p ' + Rmlint.get_options(self, paths)

    def get_benchid(self):
        return 'rmlint-paranoid'


class RmlintReplay(Rmlint):

    def get_options(self, paths):
        return '--replay /tmp/rmlint.json -T df -o summary ' + ' '.join(paths)

    def get_benchid(self):
        return 'rmlint-replay'


class OldRmlint(Program):
    website = 'https://github.com/sahib/rmlint'
    script = 'rmlint-old.sh'

    def get_binary(self):
        return 'rmlint/rmlint-old'

    def get_options(self, paths):
        return ' '.join(paths) + ' -v4'  # log output to stdout

    def compute_version(self):
        version_text = subprocess.check_output(
            'rmlint/rmlint-old --version', shell=True, stderr=subprocess.STDOUT
        ).decode('utf-8')

        match = re.search('Version (\d.\d.\d)', str(version_text))
        if match is not None:
            return ' '.join(match.groups()).strip()

        return ""

    def parse_statistics(self, dump):
        dups, sets = 0, 0
        for line in dump.splitlines():
            if line.startswith(b'ORIG'):
                sets += 1
                dups += 1

            if line.startswith(b'DUPL'):
                dups += 1

        return {
            'dupes': dups,
            'sets': sets
        }


class Dupd(Program):
    website = 'https://github.com/jvirkki/dupd'
    script = 'dupd.sh'
    stats_file = '/tmp/rmlint-bench/.dupd.stats'

    def get_binary(self):
        return 'dupd/dupd'

    def get_options(self, paths):
        try:
            os.remove(self.stats_file)
        except OSError:
            pass

        args = ' '.join(['--path ' + path for path in paths])
        return 'scan ' + args + ' --nodb --stats-file ' + self.stats_file

    def compute_version(self):
        return subprocess.check_output(
            'dupd/dupd version', shell=True
        ).decode('utf-8').strip()

    def parse_statistics(self, _):
        stats = {}

        try:
            with open(self.stats_file, 'r') as fd:
                for line in fd:
                    line = line.strip()
                    if line:
                        key, value = line.split(' ', 1)
                        stats[key] = value
            return {
                'dupes': int(stats['stats_duplicate_files']),
                'sets': int(stats['stats_duplicate_sets'])
            }
        except OSError:
            pass
        except ValueError as err:
            print('doh', err)


class Rdfind(Program):
    website = 'http://rdfind.pauldreik.se'
    script = 'rdfind.sh'
    result_file = '/tmp/rmlint-bench/.rdfind.results'

    def get_binary(self):
        return 'rdfind-1.3.4/rdfind'

    def get_options(self, paths):
        return '-ignoreempty true -removeidentinode \
            false -checksum sha1 -dryrun true \
            -outputname ' + self.result_file + ' '.join(paths)

    def compute_version(self):
        words = subprocess.check_output(
            'rdfind-1.3.4/rdfind --version', shell=True
        ).decode('utf-8')

        return words.split(' ')[-1].strip()

    def parse_statistics(self, _):
        try:
            stats = Counter()
            with open(self.result_file, 'r') as fd:
                for line in fd:
                    line = line.strip()
                    if line:
                        stats[line.split()[0]] += 1

            return {
                'dupes': stats['DUPTYPE_WITHIN_SAME_TREE'],
                'sets': stats['DUPTYPE_FIRST_OCCURRENCE']
            }
        except OSError:
            pass
        except ValueError:
            pass


class Fdupes(Program):
    website = 'http://en.wikipedia.org/wiki/Fdupes'
    script = 'fdupes.sh'

    def get_binary(self):
        return 'fdupes/fdupes'

    def get_options(self, paths):
        return '-f -q -rnH -m ' + ' '.join(paths)

    def parse_statistics(self, dump):
        dump = dump.decode('utf-8')
        match = re.match(
            '(\d+) duplicate files \(in (\d+) sets\)', dump
        )

        if match:
            return {
                'dupes': int(match.group(1)),
                'sets': int(match.group(2))
            }

    def compute_version(self):
        return subprocess.check_output(
            'fdupes/fdupes --version', shell=True
        ).decode('utf-8').strip()


class Baseline(Program):
    website = 'https://github.com/sahib/rmlint/blob/develop\
/tests/test_speed/build_scripts/baseline.sh'

    script = 'baseline.sh'

    def get_binary(self):
        return 'baseline/baseline.py'

    def get_options(self, paths):
        return ' '.join(paths)

    def compute_version(self):
        return '1.0'

    def parse_statistics(self, dump):
        return json.loads(dump.decode('utf-8'))

############################
#    DATASET GENERATORS    #
############################


class Dataset:
    __metaclass__ = ABCMeta

    def __init__(self, name):
        self.name = name

    ####################
    # ABSTRACT METHODS #
    ####################

    @abstractmethod
    def generate(self):
        pass

    @abstractmethod
    def get_paths(self):
        pass


class ExistingDataset(Dataset):

    def __init__(self, name, paths):
        Dataset.__init__(self, name)
        self.paths = paths

    def generate(self):
        pass

    def get_paths(self):
        return self.paths


def do_run(programs, dataset):
    results = {
        'metadata': {
            'path': dataset.get_paths(),
            'n_runs': CFG.n_runs,
            'n_sub_runs': CFG.n_sub_runs
        },
        'programs': {}
    }

    benchmark_json_path = os.path.join(
        CFG.output,
        'benchmark_{name}.json'.format(
            name=dataset.name.replace('/', '\\')
        )
    )

    for program in programs:
        program.install()

        data, memory_usage = program.run(dataset)
        print('>> Timing was #run: [time, cpu, dupes, sets]:')
        pprint.pprint(data)

        bench_id = program.get_benchid()
        results['programs'][bench_id] = {}
        results['programs'][bench_id]['version'] = program.version
        results['programs'][bench_id]['website'] = program.get_website()
        results['programs'][bench_id]['numbers'] = data
        results['programs'][bench_id]['memory'] = memory_usage

        print('-- Writing current benchmark results to', benchmark_json_path)
        with open(benchmark_json_path, 'w') as json_file:
            json_file.write(json.dumps(results, indent=4))


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-r", "--run", help="Run the benchmarks.",
        dest='do_run', default=False, action='store_true'
    )
    parser.add_argument(
        "-i", "--install", help="Install all competitors from source.",
        dest='do_install', default=False, action='store_true'
    )
    parser.add_argument(
        "-g", "--generate", help="Generate datasets.",
        dest='do_generate', default=False, action='store_true'
    )
    parser.add_argument(
        "-v", "--print-versions", help="Print versions of programs.",
        dest='do_print_versions', default=False, action='store_true'
    )
    parser.add_argument(
        "-d", "--dataset",
        help="Dataset to run on. Can be given multiple times.",
        dest='datasets', action='append'
    )
    parser.add_argument(
        "-p", "--program",
        help="Choose program to run. Can be given multiple times.",
        dest='programs', action='append'
    )
    return parser.parse_args()


def main():
    datasets = [
        ExistingDataset('usr', ['/usr']),
        ExistingDataset('tmp', ['/tmp']),
        ExistingDataset('tmpvar', ['/tmp', '/var']),
        ExistingDataset('usrvar', ['/usr', '/var']),
        ExistingDataset('usr_music', ['/usr', '/mnt/music']),
        ExistingDataset('home', ['/home/sahib'])
    ]

    options = parse_arguments()

    # Make specifying absolute paths to -d possible.
    for dataset_name in options.datasets:
        if dataset_name.startswith('/'):
            datasets.append(ExistingDataset(
                dataset_name,
                [dataset_name]
            ))

    if options.do_generate or options.datasets:
        for dataset in datasets:
            if dataset in options.datasets or len(options.datasets) is 0:
                dataset.generate()

    programs = [
        # (hopefully) slowest:
        # Baseline(),
        # Current:
        Rmlint(),
        RmlintParanoid(),
        # RmlintSpot(),
        # RmlintSpotParanoid(),
        Rmlint222(),
        Rmlint246(),
        Rmlint222Paranoid(),
        RmlintXXHash(),
        # RmlintSpooky(),
        # RmlintCity(),
        # RmlintMD5(),
        # RmlintMurmur(),
        RmlintReplay(),
        # Old rmlint:
        OldRmlint(),
        # Actual competitors:
        Fdupes(),
        Rdfind(),
        # Dupd()
    ]

    # Filter the `programs` list if necessary.
    if options.programs:
        options.programs = set(options.programs)
        programs = [p for p in programs if p.get_benchid() in options.programs]

    if not programs:
        print('!! No valid programs given.')
        return

    # Do the install procedure only:
    if options.do_install:
        for program in programs:
            print('++ Installing ' + program.get_name() + ':')
            program.install()

    # Print informative version if needed.
    if options.do_print_versions:
        current_path = os.getcwd()
        for program in programs:
            os.chdir(program.get_temp_dir())
            print(program.get_binary() + ':', program.compute_version())

        os.chdir(current_path)

    # Execute the actual run:
    if options.do_run:
        # Make sure the output directory exists:
        try:
            os.mkdir(CFG.output)
        except OSError:
            pass

        for dataset_name in options.datasets:
            for dataset in datasets:
                if dataset.name == dataset_name:
                    try:
                        do_run(programs, dataset)
                        break
                    except KeyboardInterrupt:
                        print(' Interrupted this run. Next dataset.')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
from parameterized import parameterized


def create_files():
    # Same size, different content.
    create_file('a', '1.a')
    create_file('b', '1.b')

    # Same size, same content.
    create_file('a', '2.a')
    create_file('a', '2.a_')

    # Different content and size
    create_file('a' * 3, '3.a')
    create_file('a' * 4, '3.a_')

    # Size group that will generate ext_cksums for all files
    create_file('b' * 2, '4.a')
    create_file('b' * 2, '4.b')
    create_file('c' * 2, '4.c')
    create_file('c' * 2, '4.d')

    # duplicate_dirs + with --write_unfinished
    create_file('x', 'dir_a/1')
    create_file('x', 'dir_b/1')


def check(data, write_cache):
    unfinished = [p['path'] for p in data if p['type'] == 'unique_file']
    dupe_files = [p['path'] for p in data if p['type'] == 'duplicate_file']
    dupe_trees = [p['path'] for p in data if p['type'] == 'duplicate_dir']

    path_in = lambda name, paths: os.path.join(TESTDIR_NAME, name) in paths

    if write_cache:
        assert len(unfinished) == 3
        assert path_in('1.b', unfinished)
        assert path_in('dir_a/1', unfinished)
        assert path_in('dir_b/1', unfinished)

    assert len(dupe_trees) == 2
    assert path_in('dir_a', dupe_trees)
    assert path_in('dir_b', dupe_trees)

    assert len(dupe_files) == 7
    assert path_in('2.a', dupe_files)
    assert path_in('2.a_', dupe_files)
    assert path_in('1.a', dupe_files)
    assert path_in('4.a', dupe_files)
    assert path_in('4.b', dupe_files)
    assert path_in('4.c', dupe_files)
    assert path_in('4.d', dupe_files)


@with_setup(usual_setup_func, usual_teardown_func)
def test_xattr_basic():
    create_files()

    for _ in range(2):
        for write_cache in True, False:
            if write_cache:
                head, *data, footer = run_rmlint('-U -D -S pa --xattr-write')
            else:
                head, *data, footer = run_rmlint('-D -S pa --xattr-read')

            check(data, write_cache)

        head, *data, footer = run_rmlint('-D -S pa --xattr-clear')


@parameterized([("", ), ("-D", )])
@with_setup(usual_setup_func, usual_teardown_func)
def test_xattr_detail(extra_opts):
    if not runs_as_root():
        # This tests need a ext4 fs which is created during the test.
        # The mount step sadly needs root priviliges.
        return

    with create_special_fs("this-is-not-tmpfs") as ext4_path:
        # Keep the checksum fixed, if we change the default we don't want to
        # break this test (although I'm sure some tests will break)
        base_options = extra_opts + " -T df -S pa -a blake2b "

        path_1 = os.path.join(ext4_path, "1")
        path_2 = os.path.join(ext4_path, "2")
        path_3 = os.path.join(ext4_path, "3")
        path_4 = os.path.join(ext4_path, "4")

        create_file("abc", path_1)
        create_file("abc", path_2)
        create_file("def", path_3)
        create_file("longer", path_4)

        head, *data, footer = run_rmlint(base_options + ' --xattr-write')
        assert len(data) == 2

        xattr_1 = must_read_xattr(path_1)
        xattr_2 = must_read_xattr(path_2)
        xattr_3 = must_read_xattr(path_3)
        assert xattr_1["user.rmlint.blake2b.cksum"] == \
                b'ba80a53f981c4d0d6a2797b69f12f6e94c212f14685ac4b74b12bb6fdbffa2d17d87c5392aab792dc252d5de4533cc9518d38aa8dbf1925ab92386edd4009923\x00'
        assert xattr_1 == xattr_2

        # no --write-unfinished given.
        assert xattr_3 == {}

        # Repeating the caching option should have no effect on the output.
        for _ in range(10):
            head, *data, footer = run_rmlint(base_options + ' --xattr')
            # one more due to the unique_file
            assert len(data) == 3

            xattr_1 = must_read_xattr(path_1)
            xattr_2 = must_read_xattr(path_2)
            xattr_3 = must_read_xattr(path_3)
            assert xattr_1["user.rmlint.blake2b.cksum"] == \
                    b'ba80a53f981c4d0d6a2797b69f12f6e94c212f14685ac4b74b12bb6fdbffa2d17d87c5392aab792dc252d5de4533cc9518d38aa8dbf1925ab92386edd4009923\x00'
            assert xattr_1 == xattr_2

            # --write-unfinished will also write the unfinished one.
            xattr_3 = must_read_xattr(path_3)
            assert xattr_3["user.rmlint.blake2b.cksum"] == \
                    b'36badf2227521b798b78d1bd43c62520a35b9b541547ff223f35f74b1168da2cd3c8d102aaee1a0cc217b601258d80151067cdee3a6352517b8fc7f7106902d3\x00'

            # unique file which was not hashed -> does not need to be touched.
            xattr_4 = must_read_xattr(path_4)
            assert xattr_4 == {}

        # Try clearing the attributes:
        head, *data, footer = run_rmlint(base_options + '--xattr-clear')
        assert len(data) == 2
        assert must_read_xattr(path_1) == {}
        assert must_read_xattr(path_2) == {}
        assert must_read_xattr(path_3) == {}
        assert must_read_xattr(path_4) == {}



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_keep_hardlinks():
    create_file('xxx', 'file_a')
    create_link('file_a', 'file_b')
    create_file('xxx', 'file_z')

    head, *data, footer = run_rmlint('--no-hardlinked -S a')
    assert data[0]["path"].endswith("file_a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_z")
    assert data[1]["is_original"] is False

    head, *data, footer = run_rmlint('--hardlinked -S a')
    assert data[0]["path"].endswith("file_a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_b")
    assert data[1]["is_original"] is False
    assert data[2]["path"].endswith("file_z")
    assert data[2]["is_original"] is False

    head, *data, footer = run_rmlint('--keep-hardlinked -S a')
    assert data[0]["path"].endswith("file_a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_b")
    assert data[1]["is_original"] is True
    assert data[2]["path"].endswith("file_z")
    assert data[2]["is_original"] is False

    head, *data, footer = run_rmlint('--keep-hardlinked -S A')
    assert data[0]["path"].endswith("file_z")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_b")
    assert data[1]["is_original"] is False
    assert data[2]["path"].endswith("file_a")
    assert data[2]["is_original"] is False


def test_keep_hardlinks_multiple_originals():
    create_file('xxx', 'a/file_a')
    create_file('xxx', 'a/file_y')
    create_dirs('b')
    create_link('a/file_a', 'b/file_b')
    create_link('a/file_y', 'b/file_z')

    search_paths = TESTDIR_NAME + '/b // ' + TESTDIR_NAME + '/a'

    head, *data, footer = run_rmlint('--no-hardlinked -S a ' + search_paths, use_default_dir=False)
    # hardlinks file_b and file_z should be ignored
    assert len(data)==2
    assert data[0]["path"].endswith("file_a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_y")
    assert data[1]["is_original"] is False

    head, *data, footer = run_rmlint('--hardlinked -k -m -S a ' + search_paths, use_default_dir=False)
    # files in folder a should both be originals because tagged
    assert len(data)==4
    assert data[0]["path"].endswith("file_a")
    assert data[0]["is_original"] is True
    assert data[1]["path"].endswith("file_y")
    assert data[1]["is_original"] is True
    assert data[2]["path"].endswith("file_b")
    assert data[2]["is_original"] is False
    assert data[3]["path"].endswith("file_z")
    assert data[3]["is_original"] is False

    head, *data, footer = run_rmlint('--keep-hardlinked -k -m -S a ' + search_paths, use_default_dir=False)
    # files in folder a are tagged so should both be preserved;
    # files in folder b are hardlinks of the two originals so should also be preserved
    # therefore all files are originals and so don't get reported
    assert len(data)==0



#!/usr/bin/env python3
# encoding: utf-8
import os
import json

from nose import with_setup
from contextlib import contextmanager

from tests.utils import *


@contextmanager
def assert_exit_code(status_code):
    """
    Assert that the with block yields a subprocess.CalledProcessError
    with a certain return code. If nothing is thrown, status_code
    is required to be 0 to survive the test.
    """
    try:
        yield
    except subprocess.CalledProcessError as exc:
        assert exc.returncode == status_code
    else:
        # No exception? status_code should be fine.
        assert status_code == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_files():
    path_a = create_file('1234', 'a')
    path_b = create_file('1234', 'b')

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False
        )

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '-p', '--equal', path_a, path_b,
            use_default_dir=False
        )

    # If --equal finds that both files are not equal,
    # it would return 1 as exit code which would cause
    # a CalledProcessError in run_rmlint
    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')

    path_c = create_file('1234', 'c')
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b, path_c,
            use_default_dir=False
        )

    path_d = create_file('diff', 'd')
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b, path_d, path_c,
            use_default_dir=False
        )


@with_setup(usual_setup_func, usual_teardown_func)
def test_no_arguments():
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal',
            use_default_dir=False
        )


@with_setup(usual_setup_func, usual_teardown_func)
def test_one_arguments():
    path = create_file('1234', 'a')
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path,
            use_default_dir=False
        )

    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path, "//",
            use_default_dir=False
        )

    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', "//", path,
            use_default_dir=False
        )


@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_directories():
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = os.path.dirname(create_file('xxx', 'dir_b/x'))

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False
        )

    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '-p', '--equal', path_a, path_b,
            use_default_dir=False
        )


@with_setup(usual_setup_func, usual_teardown_func)
def test_dir_and_file():
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = create_file('xxx', 'x')

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(1):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )


# Regression test for Issue #233
@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_hidden_dirs():
    path_a = os.path.dirname(create_file('xxx', 'dir_a/x'))
    path_b = os.path.dirname(create_file('xxx', '.dir_b/.x'))

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )


# Regression test for Issue #234
@with_setup(usual_setup_func, usual_teardown_func)
def test_equal_empty_files_or_other_lint():
    path_a = create_file('', 'x')
    path_b = create_file('', 'y')

    # This should fail since we should not mix directories with files,
    # even if they have the same content.
    with assert_exit_code(0):
        head, *data, footer = run_rmlint(
            '--equal', path_a, path_b,
            use_default_dir=False,
        )



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from nose.plugins.attrib import attr
from tests.utils import *

from itertools import permutations, combinations


PATHS = ['b_dir/', 'a_dir/', 'c_dir/']


def path_index(file_path):
    for idx, path in enumerate(PATHS):
        if path in file_path:
            return idx
    assert False


def path_depth(file_path):
    return len([c for c in file_path if c == '/'])

def in_order(a, b, increasing):
    return (a < b) if increasing else (b < a)

# dispatcher for comparison tests
def validate_order(data, tests):
    testfuncs = {
        'a': lambda p: os.path.basename(p['path']).lower(),
        'm': lambda p: p['mtime'],
        'p': lambda p: path_index(p['path']),
        'd': lambda p: path_depth(p['path']),
        'l': lambda p: len(os.path.basename(p['path']))
    }

    for a, b in combinations(data, 2):
        for test in tests:
            cmp_a, cmp_b = (testfuncs[test.lower()](e) for e in [a, b])

            # Equal? Go to next criterion.
            if cmp_a == cmp_b:
                continue

            # special handling for mtime rounding
            if (test.lower()=='m' and abs(cmp_a-cmp_b) < 0.00000001):
                continue

            if in_order(cmp_a, cmp_b, test.islower()):
                # order is correct; ignore any remaining (less important) criteria
                break

            # Order is incorrect.
            assert False


@attr('slow')
@with_setup(usual_setup_func, usual_teardown_func)
def test_sorting():
    # create some dupes with different PATHS, names and mtimes:
    create_file('xxx', PATHS[1] + 'a', mtime='2004-02-29  16:21:42.4')
    create_file('xxx', PATHS[0] + 'c', mtime='2004-02-29  16:21:42.6')
    create_file('xxx', PATHS[2] + 'B', mtime='2004-02-29  16:21:43.1')
    create_file('xxx', PATHS[2] + 'b', mtime='2004-02-29  16:21:43.0')
    create_file('xxx', PATHS[1] + 'c', mtime='2004-02-29  16:21:41.5')
    create_file('xxx', PATHS[2] + 'c', mtime='2004-02-29  16:21:41.0')

    joiner = ' ' + TESTDIR_NAME + '/'
    search_paths = joiner + joiner.join(PATHS)

    opts = 'ampdl'
    all_opts = opts + opts.upper()

    combos = []
    is_legal_combo = lambda x: len(x) == len(set(x.lower()))

    # Limit to 3-tuple combinations. 4-5-tuple combinations (just do +1)
    # are possible if you have enough time (=> ~(10! / 2) tests).
    for n_terms in range(1, len(opts) - 1):
        combos += filter(
            is_legal_combo,
            (''.join(p) for p in permutations(all_opts, n_terms))
        )

    for combo in combos:
        combo_str = '-S ' + combo
        head, *data, footer = run_rmlint(combo_str + search_paths, use_default_dir=False)
        assert len(data) == 6

        validate_order(data, combo)

@with_setup(usual_setup_func, usual_teardown_func)
def test_sort_by_outlyer():
    create_file('xxx', 'a/foo')
    create_file('xxx', 'b/foo')

    create_link('a/foo', 'b/foo-from-a')
    create_link('b/foo', 'b/foo-copy-1')
    create_link('b/foo', 'b/foo-copy-2')

    head, *data, footer = run_rmlint(
        "-S O {t}/b".format(t=TESTDIR_NAME), use_default_dir=False
    )
    assert data[0]['path'].endswith('b/foo-from-a')

    head, *data, footer = run_rmlint('-S OHa')
    assert data[0]['path'].endswith('b/foo')


# See this issue for more information.
# https://github.com/sahib/rmlint/issues/196
#
# Testsetup by "Awerick"
@with_setup(usual_setup_func, usual_teardown_func)
def test_sort_by_outlyer_hardcore():
    for suffix in 'ABCD':
        create_file('xxx', 'inside/foo' + suffix)

    create_dirs('outside')

    # fooA:
    for i in range(2, 6):
        create_link('inside/fooA', 'inside/fooA_link_' + str(i))

    # fooB:
    for i in range(2, 5):
        create_link('inside/fooB', 'inside/fooB_link_' + str(i))
    create_link('inside/fooB', 'outside/fooB_link_5')

    # fooC:
    create_link('inside/fooC', 'inside/fooC_link_2')
    create_link('inside/fooC', 'outside/fooC_link_3')
    create_link('inside/fooC', 'outside/fooC_link_4')

    # fooD:
    create_link('inside/fooD', 'outside/fooD_link_2')
    create_link('inside/fooD', 'outside/fooD_link_3')

    def run_inside_job(crit):
        head, *data, footer = run_rmlint(
            "-S {c} {t}/inside".format(
                c=crit,
                t=TESTDIR_NAME,
            ), use_default_dir=False
        )

        return data[0]['path']

    assert run_inside_job('Ha').endswith('inside/fooA')
    assert run_inside_job('HOa').endswith('inside/fooB')
    assert run_inside_job('OHa').endswith('inside/fooC')
    assert run_inside_job('OA').endswith('inside/fooD')



@with_setup(usual_setup_func, usual_teardown_func)
def test_sort_by_regex():
    create_file('xxx', 'aaaa')
    create_file('xxx', 'aaab')
    create_file('xxx', 'b')
    create_file('xxx', 'c')
    create_file('xxx', '1/c')
    create_file('xxx', 'd')

    head, *data, footer = run_rmlint("-S 'r<1/c>x<d$>a'")

    paths = [p['path'] for p in data]

    assert paths[0].endswith('1/c')
    assert paths[1].endswith('d')
    assert paths[2].endswith('aaaa')
    assert paths[3].endswith('aaab')
    assert paths[4].endswith('b')
    assert paths[5].endswith('c')


@with_setup(usual_setup_func, usual_teardown_func)
def test_sort_by_regex_bad_input():
    create_file('xxx', 'aaaa')
    create_file('xxx', 'aaab')

    # Should work:
    run_rmlint("-S '{}'".format('r<.>' * 8))

    # More than 8 is bad:
    try:
        run_rmlint("-S '{}'".format('r<.>' * 9))
        assert False
    except subprocess.CalledProcessError:
        pass

    # Empty patterns are sill also:
    try:
        run_rmlint("-S 'r<>'")
        assert False
    except subprocess.CalledProcessError:
        pass

    # A bad regex is bad too:
    try:
        run_rmlint("-S 'r<*>'")
        assert False
    except subprocess.CalledProcessError:
        pass



#!/usr/bin/env python3
# encoding: utf-8

from nose import with_setup
from tests.utils import *
from nose.plugins.attrib import attr
from parameterized import parameterized

INCREMENTS = [4096, 1024, 1, 20000]

def streaming_compliance_check(patterns):
    # a valid hash function streaming function should satisfy hash('a', 'b', 'c') == hash('abc')

    a = create_file('1' * 10000, 'a')

    algos = []
    for pattern in patterns:
        algos += [algo for algo in CKSUM_TYPES if pattern in algo]

    cmd = './rmlint --hash --increment {increment} --algorithm {algo} {path}'

    for algo in algos:
        command = cmd.format(increment=INCREMENTS[0], algo=algo, path=a)
        output0 = subprocess.check_output(command.split())
        for increment in INCREMENTS[1:]:
            command = cmd.format(increment=increment, algo=algo, path=a)
            output = subprocess.check_output(command.split())
            if(output!=output0):
                assert False, "{} fails streaming test with increment {}".format(algo, increment)
                break

@parameterized([
        'murmur',
        'metro',
        ['glib:', 'md5', 'sha1', 'sha256', 'sha512'],
        'sha3',
        'blake',
        'xxhash',
        'highway'
        ])
def test_hash_function(*pat):
    if(len(pat)==1):
        streaming_compliance_check(pat)
    else:
        streaming_compliance_check(pat[1:])




#!/usr/bin/env python3
# encoding: utf-8
from nose.plugins.attrib import attr
from nose import with_setup
from tests.utils import *

import os
import subprocess


def set_mtime(path, mtime):
    full_path = os.path.join(TESTDIR_NAME, path)
    subprocess.call(['touch', '-m', '-d', str(mtime), full_path])


@with_setup(usual_setup_func, usual_teardown_func)
def test_consider_mtime():
    create_file('xxx', 'a')
    create_file('xxx', 'b')
    create_file('xxx', 'c')
    create_file('xxx', 'd')

    set_mtime('a', '2004-02-29  16:21:42')
    set_mtime('b', '2004-02-29  16:21:42')
    set_mtime('c', '2004-02-29  16:21:44')
    set_mtime('d', '2004-02-29  16:21:45')

    head, *data, footer = run_rmlint('--mtime-window=-1')
    assert len(data) == 4
    assert footer['total_files'] == 4
    assert footer['total_lint_size'] == 9
    assert footer['duplicates'] == 3
    assert footer['duplicate_sets'] == 1

    head, *data, footer = run_rmlint('--mtime-window=0')
    assert len(data) == 2
    assert footer['total_files'] == 4
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1
    assert footer['duplicate_sets'] == 1

    head, *data, footer = run_rmlint('--mtime-window=+1')
    assert len(data) == 4
    assert footer['total_files'] == 4
    assert footer['total_lint_size'] == 6  # two originals.
    assert footer['duplicates'] == 2
    assert footer['duplicate_sets'] == 2

    head, *data, footer = run_rmlint('--mtime-window=+2')
    assert len(data) == 4   # '2' also chains up to d from c.
    assert footer['total_files'] == 4
    assert footer['total_lint_size'] == 9
    assert footer['duplicates'] == 3
    assert footer['duplicate_sets'] == 1


@with_setup(usual_setup_func, usual_teardown_func)
def test_consider_mtime_subsecond():
    create_file('xxx', 'a')
    create_file('xxx', 'b')

    set_mtime('a', '2004-02-29  16:21:42.00')
    set_mtime('b', '2004-02-29  16:21:43.99')

    head, *data, footer = run_rmlint('--mtime-window=1.9')
    assert len(data) == 0

    head, *data, footer = run_rmlint('--mtime-window=2.0')
    assert len(data) == 2

    set_mtime('a', '2004-02-29  16:21:42.00')
    set_mtime('b', '2004-02-29  16:21:42.99')

    head, *data, footer = run_rmlint('--mtime-window=0')
    assert len(data) == 0

@with_setup(usual_setup_func, usual_teardown_func)
def test_consider_mtime_fail_by_association():
    create_file('xxx', 'a')
    create_file('yyy', 'b')
    create_file('xxx', 'c')

    set_mtime('a', '2004-02-29  16:21:42')
    set_mtime('b', '2004-02-29  16:21:44')
    set_mtime('c', '2004-02-29  16:21:46')

    head, *data, footer = run_rmlint('--mtime-window=3')

    assert len(data) == 0
    assert footer['total_files'] == 3
    assert footer['total_lint_size'] == 0
    assert footer['duplicates'] == 0
    assert footer['duplicate_sets'] == 0

@with_setup(usual_setup_func, usual_teardown_func)
def test_mtime_and_unmatched_basenames():
    create_file('xxx', 'dir1/a')
    create_file('xxx', 'dir1/c')
    create_file('xxx', 'dir2/a')

    create_file('yyy', 'dir1/b')
    create_file('yyy', 'dir2/b')
    create_file('yyy', 'dir2/c')

    set_mtime('dir1/a', '2004-02-29  16:21:42')
    set_mtime('dir1/c', '2004-02-29  16:21:44')
    set_mtime('dir2/a', '2004-02-29  16:21:48')

    set_mtime('dir1/b', '2004-02-29  16:21:46')
    set_mtime('dir2/b', '2004-02-29  16:21:48')
    set_mtime('dir2/c', '2004-02-29  16:21:50')

    head, *data, footer = run_rmlint('--mtime-window=3 --unmatched-basename -S m')

    assert len(data) == 2
    assert footer['total_files'] == 6
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1
    assert footer['duplicate_sets'] == 1



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


def create_set():
    for suffix in 'abc':
        create_file('x' * 2048, 'big' + suffix)
        create_file('x' * 1024, 'middle' + suffix)
        create_file('x' * 512, 'small' + suffix)


@with_setup(usual_setup_func, usual_teardown_func)
def test_valid():
    create_set()

    # Scalar:
    *_, footer = run_rmlint('--size 0')
    assert footer['duplicates'] == 6
    *_, footer = run_rmlint('--size 1024')
    assert footer['duplicates'] == 4
    *_, footer = run_rmlint('--size 2048')
    assert footer['duplicates'] == 2
    *_, footer = run_rmlint('--size 2049')
    assert footer['duplicates'] == 0

    # Ranges:
    *_, footer = run_rmlint('--size 1024-2048')
    assert footer['duplicates'] == 4

    *_, footer = run_rmlint('--size 0-1024')
    assert footer['duplicates'] == 4

    *_, footer = run_rmlint('--size 2048-2048')
    assert footer['duplicates'] == 2

    *_, footer = run_rmlint('--size 2K-2KB')
    assert footer['duplicates'] == 2

    *_, footer = run_rmlint('--size 2K-2KB')
    assert footer['duplicates'] == 2

    *_, footer = run_rmlint('--size 18446744073709551615-18446744073709551615')
    assert footer['duplicates'] == 0

    *_, footer = run_rmlint('--size 1-18446744073709551615')
    assert footer['duplicates'] == 6

@with_setup(usual_setup_func, usual_teardown_func)
def test_invalid():
    create_set()

    def trigger(*args):
        try:
            run_rmlint(*args)
        except subprocess.CalledProcessError:
            pass
        else:
            print(args, 'did not trigger an error exit.')
            assert False

    # Not a valid range:
    trigger('--size \-\-17')

    # max < min
    trigger('--size 10-9')

    # double min
    trigger('--size 10--10')

    # overflow by one.
    trigger('--size 0-18446744073709551616')

    # overflow by factor.
    trigger('--size 0-18446744073709551615M')

    # overflow by fraction.
    trigger('--size 0-18446744073709551615.1')



@with_setup(usual_setup_func, usual_teardown_func)
def test_replay_size():
    create_file('', 'empty1')
    create_file('', 'empty2')
    create_file('xxx', 'a/xxx')
    create_file('xxx', 'b/xxx')
    create_file('yyy', 'a/yyy')
    create_file('yyy', 'b/yyy')
    create_testdir('empty_dir')

    replay_path = '/tmp/replay.json'
    head, *data, footer = run_rmlint('-o json:{p}'.format(
        p=replay_path
    ))

    assert len(data) == 7
    assert [e["type"] for e in data] == \
           ["emptydir"] + (["emptyfile"] * 2) + (["duplicate_file"] * 4)

    head, *data, footer = run_rmlint('--replay {p} --size 1-10B'.format(
        p=replay_path
    ))

    assert [e["type"] for e in data] == \
           ["emptydir"] + (["emptyfile"] * 2) + (["duplicate_file"] * 4)



#!/usr/bin/env python3
# encoding: utf-8
#!/usr/bin/env python

from distutils.core import setup
from distutils.command.install import install

import os
import sys
import subprocess


def read_version():
    with open('../.version', 'r') as handle:
        version_string = handle.read()

    return version_string.strip()


def get_prefix():
    # distutils apparently has no sane way to get the install prefix early.
    # (I realize that this is a stupid hack to do something obvious)
    for idx, arg in enumerate(sys.argv):
        if arg == '--user':
            return os.path.expanduser('~/.local')

        if arg.startswith('--prefix'):
            if '=' in arg:
                _, path = arg.split('=', 1)
                return path
            else:
                return sys.argv[idx + 1]

    return '/usr'


PREFIX = get_prefix()


class PrePlusPostInstall(install):
    def run(self):
        # Compile the resource bundle freshly
        print('==> Compiling resource bundle')

        if os.access('shredder/resources/shredder.gresource', os.R_OK):
            print('==> Using existing. Lucky boy.')
        else:
            print('==> Calling glib-compile-resources')
            try:
                subprocess.call([
                    'glib-compile-resources',
                    'shredder/resources/shredder.gresource.xml',
                    '--sourcedir',
                    'shredder/resources'
                ])
            except subprocess.CalledProcessError as err:
                print('==> Failed :(')

        # Run the usual distutils install routine:
        install.run(self)

        # Make sure the schema file is updated.
        # Otherwise the gui will trace trap.
        print('==> Compiling GLib Schema files')

        try:
            subprocess.call([
                'glib-compile-schemas',
                os.path.join(PREFIX, 'share/glib-2.0/schemas')
            ])
        except subprocess.CalledProcessError as err:
            print('==> Could not update schemas: ', err)
            print('==> Please run the following manually:\n')
            print('    sudo glib-compile-schemas {prefix}'.format(
                prefix=os.path.join(PREFIX, 'share/glib-2.0/schemas')
            ))
        else:
            print('==> OK!')


setup(
    name='Shredder',
    version=read_version(),
    description='A gui frontend to rmlint',
    long_description='A graphical user interface to rmlint using GTK+',
    author='Christopher Pahl',
    author_email='sahib@online.de',
    url='https://rmlint.rtfd.org',
    license='GPLv3',
    platforms='any',
    cmdclass={'install': PrePlusPostInstall},
    packages=['shredder', 'shredder.views'],
    package_data={'': [
        'resources/*.gresource'
    ]},
    data_files=[
        (
            os.path.join(PREFIX, 'share/icons/hicolor/scalable/apps'),
            ['shredder/resources/shredder.svg']
        ),
        (
            os.path.join(PREFIX, 'share/glib-2.0/schemas'),
            ['shredder/resources/org.gnome.Shredder.gschema.xml']
        ),
        (
            os.path.join(PREFIX, 'share/applications'),
            ['shredder.desktop']
        ),
    ]
)



#!/usr/bin/env python
# encoding: utf-8

"""
Main view of Shredder.

Shows the chart and a treeview of suspicious files.
"""

# Stdlib:
import logging

# External:
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import GObject

# Internal:
from shredder.util import View, IconButton, NodeState
from shredder.util import MultipleChoiceButton, scrolled
from shredder.chart import ChartStack
from shredder.tree import PathTreeView, PathTreeModel, Column
from shredder.runner import Runner


LOGGER = logging.getLogger('runview')
RENDER_CHOICES = ['All', 'Selected', 'Filtered']


class ResultActionBar(Gtk.ActionBar):
    """Down right bar with the controls"""
    __gsignals__ = {
        'generate-all-script': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'generate-filtered-script': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'generate-selection-script': (GObject.SIGNAL_RUN_FIRST, None, ())
    }

    def __init__(self, view):
        Gtk.ActionBar.__init__(self)

        left_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        left_box.get_style_context().add_class("linked")
        self.pack_start(left_box)

        self.refresh_button = IconButton('view-refresh-symbolic')
        self.settings_button = IconButton('system-run-symbolic')

        self.refresh_button.connect(
            'clicked', lambda _: view.app_window.views['runner'].rerun()
        )
        self.settings_button.connect(
            'clicked', lambda _: view.app_window.views.switch('settings')
        )

        left_box.pack_start(self.refresh_button, False, False, 0)
        left_box.pack_start(self.settings_button, False, False, 0)

        right_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        right_box.get_style_context().add_class(
            Gtk.STYLE_CLASS_LINKED
        )

        self.script_btn = IconButton(
            'emblem-documents-symbolic', 'Render script from'
        )
        self.script_btn.connect(
            'clicked', self.on_generate_script
        )

        self.script_type_btn = MultipleChoiceButton(
            RENDER_CHOICES, 'All', 'All'
        )
        self.script_type_btn.set_relief(Gtk.ReliefStyle.NORMAL)

        right_box.pack_start(self.script_btn, True, True, 0)
        right_box.pack_start(self.script_type_btn, True, False, 0)

        self.pack_end(right_box)
        self.set_sensitive(False)

    def set_choice(self, choice):
        """Set the current choice. Might be one of RENDER_CHOICES"""
        if choice not in RENDER_CHOICES:
            raise ValueError("Bad choice for button: " + choice)

        self.script_type_btn.set_selected_choice(choice)

    def on_generate_script(self, _):
        """Called when the left side of the compound button was clicked."""
        choice = self.script_type_btn.get_selected_choice().lower()

        if choice == 'all':
            self.emit('generate-all-script')
        elif choice == 'filtered':
            self.emit('generate-filtered-script')
        elif choice == 'selected':
            self.emit('generate-selection-script')
        else:
            LOGGER.error('Bug: bad choice selection: %s', choice)

    def set_sensitive(self, mode):
        """Set the gen-script button (non)-sensitive and (non)-suggested"""
        btn_ctx = self.script_btn.get_style_context()
        type_ctx = self.script_type_btn.get_style_context()

        if mode:
            btn_ctx.add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
            type_ctx.add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
        else:
            btn_ctx.remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
            type_ctx.remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)

        self.script_btn.set_sensitive(mode)
        self.script_type_btn.set_sensitive(mode)

    def is_sensitive(self):
        return self.script_btn.is_sensitive()


class RunnerView(View):
    """Main action View.

    Public attributes:

        - script: A Script instance.
        - runner: The current run-instance.
        - model: The data.
    """
    def __init__(self, app):
        View.__init__(self, app, 'Running…')

        # Public: The runner.
        self.runner = None

        self.last_paths = []

        # Disable scrolling for the main view:
        self.scw.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.NEVER)

        # Public flag for checking if the view is still
        # in running mode (thus en/disabling certain features)
        self.is_running = False

        self._script_generated = False
        self._is_filtered = False

        self.model = PathTreeModel([])

        self.treeview = PathTreeView()
        self.treeview.set_model(self.model)
        self.treeview.get_selection().connect(
            'changed',
            self.on_selection_changed
        )

        self.group_treeview = PathTreeView()
        self.group_treeview.set_vexpand(True)
        self.group_treeview.set_valign(Gtk.Align.FILL)

        # This is needed to make sure operations on the one update
        # the other. Interally the same nodes are updated, but it has
        # to be made sure that the models get updated.
        self.group_treeview.set_twin(self.treeview)
        self.treeview.set_twin(self.group_treeview)

        group_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        group_box.pack_start(scrolled(self.group_treeview), True, True, 0)
        group_box.pack_start(Gtk.HSeparator(), False, False, 0)

        self.group_revealer = Gtk.Revealer()
        self.group_revealer.set_vexpand(True)
        self.group_revealer.set_valign(Gtk.Align.FILL)
        self.group_revealer.add(group_box)
        self.group_revealer.set_no_show_all(True)
        self.group_revealer.set_size_request(-1, 70)

        for column in self.treeview.get_columns():
            column.connect(
                'clicked',
                lambda _: self.rerender_chart()
            )

        self.chart_stack = ChartStack()
        self.actionbar = ResultActionBar(self)
        self.actionbar.set_valign(Gtk.Align.END)
        self.actionbar.set_halign(Gtk.Align.FILL)

        # Right part of the view
        stats_box = Gtk.Paned(orientation=Gtk.Orientation.VERTICAL)
        stats_box.pack1(self.group_revealer, True, True)
        stats_box.pack2(self.chart_stack, True, True)
        stats_box.props.position = 200

        # Separator container for separator|chart (could have used grid)
        separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
        right_pane = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        right_pane.pack_start(separator, False, False, 0)
        right_pane.pack_start(stats_box, True, True, 0)
        right_pane.set_size_request(100, -1)

        scw = scrolled(self.treeview)
        scw.set_size_request(200, -1)

        paned = Gtk.Paned(orientation=Gtk.Orientation.HORIZONTAL)
        paned.set_vexpand(True)
        paned.set_valign(Gtk.Align.FILL)
        paned.pack1(scw, True, True)
        paned.pack2(right_pane, True, True)
        paned.props.position = 720
        paned.set_hexpand(True)

        grid = Gtk.Grid()
        grid.attach(paned, 0, 0, 1, 1)
        grid.attach(self.actionbar, 0, 1, 1, 1)

        self.add(grid)

        self.search_entry.connect(
            'search-changed', self.on_search_changed
        )

        self.actionbar.connect(
            'generate-all-script', self.on_generate_script
        )
        self.actionbar.connect(
            'generate-filtered-script', self.on_generate_filtered_script
        )
        self.actionbar.connect(
            'generate-selection-script', self.on_generate_selection_script
        )

        self._menu = None

    def reset(self):
        """Reset internally to freshly initialized."""
        self.is_running = False
        self._script_generated = False
        self.runner = None
        self.last_paths = []

        self.chart_stack.set_visible_child_name(ChartStack.LOADING)
        self.actionbar.set_sensitive(False)

    def trigger_run(self, untagged_paths, tagged_paths):
        """Trigger a new run on all paths in `paths`"""
        # Remember last paths for rerun()
        self.reset()
        self.last_paths = (untagged_paths, tagged_paths)

        # Make sure it looks busy:
        self.sub_title = 'Running…'

        # Fork off the rmlint process:
        self.runner = Runner(self.app.settings, untagged_paths, tagged_paths)
        self.runner.connect('lint-added', self.on_add_elem)
        self.runner.connect('process-finished', self.on_process_finish)
        self.runner.run()

        # Make sure the previous run is not visible anymore:
        self.model = PathTreeModel(untagged_paths + tagged_paths)
        self.treeview.set_model(self.model)

        # Indicate that we're in a fresh run:
        self.is_running = True
        self.show_progress(0)

    def rerun(self):
        """Rerun with last given paths."""
        self.trigger_run(*self.last_paths)

    ###########################
    #     SIGNAL CALLBACKS    #
    ###########################

    def on_search_changed(self, entry):
        """Called once the user entered a new query."""
        text = entry.get_text()

        if text != "":
            self._is_filtered, choice = True, "Filtered"
        else:
            self._is_filtered, choice = False, "All"

        self.actionbar.set_choice(choice)

        sub_model = self.model.filter_model(text)
        if sub_model is not self.treeview.get_model():
            self.chart_stack.render(sub_model.trie.root)
            self.treeview.set_model(sub_model)

    def on_add_elem(self, runner):
        """Called once the runner found a new element."""
        elem = runner.element
        self.model.add_path(elem['path'], Column.make_row(elem))

        # Decide how much progress to show (or just move a bit)
        tick = (elem.get('progress', 0) / 100.0) or None
        self.show_progress(tick)

    def on_process_finish(self, _, error_msg):
        """Called once self.runner finished running."""
        # Make sure we end up at 100% progress and show
        # the progress for a short time after (for the nice cozy feeling)
        LOGGER.info('`rmlint` finished.')
        self.show_progress(100)
        GLib.timeout_add(300, self.hide_progress)
        GLib.timeout_add(350, self.treeview.expand_all)

        self.sub_title = 'Finished scanning.'

        if error_msg is not None:
            self.app_window.show_infobar(
                error_msg, message_type=Gtk.MessageType.WARNING
            )

        GLib.timeout_add(1000, self.on_delayed_chart_render, -1)

    def on_delayed_chart_render(self, last_size):
        """Called after a short delay to reduce chart redraws."""
        model = self.treeview.get_model()
        current_size = len(model)

        if current_size == last_size:
            # Come back later:
            return False

        if model.trie.has_leaves():
            self.chart_stack.set_visible_child_name(ChartStack.CHART)
            self.rerender_chart()
            self.actionbar.set_sensitive(True)
        else:
            self.chart_stack.set_visible_child_name(ChartStack.EMPTY)

        GLib.timeout_add(1500, self.on_delayed_chart_render, current_size)

        return False

    def rerender_chart(self):
        """Re-render the chart from the current model root."""
        LOGGER.info('Refreshing chart.')
        model = self.treeview.get_model()
        self.chart_stack.render(model.trie.root)

    def on_view_enter(self):
        """Called when the view enters sight."""
        GLib.idle_add(
            lambda: self.app_window.views.go_right.set_sensitive(
                self._script_generated
            )
        )

    def on_view_leave(self):
        """Called when the view leaves sight."""
        self.app_window.views.go_right.set_sensitive(True)

    def on_selection_changed(self, _):
        """Called when the user clicks a specific row."""
        node = self.treeview.get_selected_node()

        # Nothing selected:
        if node is None:
            # "Filtered" trumps "All"
            if not self._is_filtered:
                self.actionbar.set_choice("All")
            return

        self.actionbar.set_choice("Selected")

        if not node.children:
            # It is a single file.
            # Show a chart containing all twins of this file.
            # This is helpful to see quickly where those lie.

            cksum = node[Column.CKSUM]
            group = self.model.trie.group(cksum)

            paths = []
            if len(self.last_paths) > 1:
                paths = self.last_paths[0] + self.last_paths[1]

            group_model = PathTreeModel(paths)

            for twin_node in group or []:
                group_model.add_path(
                    twin_node.build_path(),
                    twin_node.row,
                    immediately=True
                )

            self.group_treeview.set_model(group_model)
            self.group_revealer.show()
            self.group_revealer.get_child().show_all()
            self.group_revealer.set_reveal_child(True)
            self.chart_stack.render(group_model.trie.root)
        else:
            self.group_revealer.hide()
            self.group_revealer.set_reveal_child(False)
            self.chart_stack.render(node)

    def _generate_script(self, trie, nodes):
        """Do the actual generation work, starting at `node` in `trie`."""
        self._script_generated = True

        path_to_is_original = {}
        for node in nodes:
            for ch in trie.iterate(node=node):
                if not ch.is_leaf:
                    continue

                path_to_is_original[ch.build_path()] = \
                    NodeState.should_keep(ch[Column.TAG])

        self.runner.replay(path_to_is_original)
        self.app_window.views.go_right.set_sensitive(True)
        self.app_window.views.switch('editor')

    def on_generate_script(self, _):
        """Generate the full script."""
        self._generate_script(self.model.trie, [self.model.trie.root])

    def on_generate_filtered_script(self, _):
        """Generate the script with only the visible content."""
        model = self.treeview.get_model()
        self._generate_script(model.trie, [model.trie.root])

    def on_generate_selection_script(self, _):
        """Generate the script only from the current selected dir or files."""
        model = self.treeview.get_model()
        selected_nodes = self.treeview.get_selected_nodes()

        if not selected_nodes:
            LOGGER.info('Nothing selected to make script from.')
            return

        self._generate_script(model.trie, selected_nodes)

    def on_default_action(self):
        """Called on Ctrl-Enter"""
        if self.actionbar.is_sensitive():
            trie = self.model.trie
            self._generate_script(trie, trie.root)



#!/usr/bin/env python
# encoding: utf-8

# Stdlib:
import os
import sys
import json
import glob
import argparse
import statistics

# External:
import pygal


# from pygal.style import LightSolarizedStyle as Style
from pygal.style import DefaultStyle as Style


VALID_ATTRS = {
    'timing': 0,
    'cpu_usage': 1,
    'peakmem': 2,
    'dupes': 3,
    'sets': 4
}

Style.background = '#FFFFFF'

CONFIG = pygal.Config()
CONFIG.human_readable = True
CONFIG.tooltip_fancy_mode = True
CONFIG.style = Style


def unpack(chart, data, bench_name, add_x_labels=True):
    chart.title = '{name} comparison on {path}'.format(
        name=bench_name,
        path=data['metadata']['path']
    )

    if bench_name:
        chart.y_title = 'Averaged {name} over {rn} runs'.format(
            rn=data['metadata']['n_sub_runs'],
            name=bench_name
        )

    n_runs = data['metadata']['n_runs']
    if add_x_labels:
        labels = ['Run #' + str(i + 1) for i in range(n_runs)]
        chart.x_labels = labels + ['Average']

    for program in sorted(data['programs']):
        result = data['programs'][program]

        if len(result["numbers"]) < n_runs:
            continue

        points = []
        for key in [str(run + 1) for run in range(n_runs)] + ['average']:
            points.append(result['numbers'][key])

        yield program, result, points


def format_tooltip(program, version):
    return '{name} ({ver})'.format(
        name=program,
        ver=version
    )


def _plot_generic(data, chart, name, key):
    for program, metadata, points in unpack(chart, data, name):
        numbers = []

        for point in points:
            numbers.append({
                'value': round(key(point), 3),
                'label': format_tooltip(
                    program,
                    metadata['version']
                ),
                'xlink': metadata.get('website')
            })

        chart.add(program, numbers)


def plot_memory(data):
    chart = pygal.StackedBar(CONFIG)
    _plot_generic(
        data, chart, 'Peakmem', lambda p: p[VALID_ATTRS['peakmem']]
    )

    return chart.render(is_unicode=True)


def plot_timing(data):
    chart = pygal.Bar(CONFIG, logarithmic=True)
    _plot_generic(
        data, chart, 'Timing', lambda p: p[VALID_ATTRS['timing']]
    )

    return chart.render(is_unicode=True)


def plot_cpu_usage(data):
    chart = pygal.Bar(CONFIG)
    _plot_generic(
        data, chart, 'CPU usage', lambda p: p[VALID_ATTRS['cpu_usage']]
    )

    return chart.render(is_unicode=True)


def plot_found_results(data):
    chart = pygal.Bar(CONFIG)

    results = []
    unpacked = unpack(chart, data, 'Found results', add_x_labels=True)
    for program, metadata, points in unpacked:
        average = points[-1]

        results.append(([
            average[VALID_ATTRS['dupes']],
            average[VALID_ATTRS['sets']],
            #statistics.variance([p[2] for p in points[:-1]]),
            #statistics.variance([p[3] for p in points[:-1]])
        ], program))

    labels = [
        'Dupes found',
        'Set of dupes',
        'Dupevariance between runs',
        'Setvariance between runs'
    ]

    results.sort()

    for point, program in results:
        chart.add(
            program, [{
                'value': value,
                'label': '{v} {l}'.format(
                    v=value, l=labels[idx]
                )
            } for idx, value in enumerate(point)]
        )

    chart.x_labels = ['Duplicates', 'Originals']
    return chart.render_table(style=True, total=False)


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'input_dir', help='Directory with bench files'
    )
    parser.add_argument(
        'output_dir', help='Where to store the plots', nargs='?'
    )

    return parser.parse_args()


def guess_output_dir(input_dir):
    if input_dir.startswith('bench-'):
        _, stamp = input_dir.split('-', 1)
        return 'plot-' + stamp

    return 'plot-output'


def main():
    options = parse_arguments()

    if not options.output_dir:
        options.output_dir = guess_output_dir(options.input_dir)

    os.makedirs(options.output_dir, exist_ok=True)

    PLOT_FUNCS = {
        'timing': (plot_timing, False),
        'memory': (plot_memory, False),
        'cpu_usage': (plot_cpu_usage, False),
        'found_items': (plot_found_results, True)
    }

    for path in glob.glob(os.path.join(options.input_dir, '*.json')):
        with open(path, 'r') as handle:
            data = json.loads(handle.read())
            for attr, (plot_func, is_table) in PLOT_FUNCS.items():
                svg_or_table = plot_func(data)
                suffix = '.html' if is_table else '.svg'

                output_path = os.path.join(
                    options.output_dir,
                    attr + '-' + os.path.basename(path) + suffix
                )

                print('Writing:', output_path)
                with open(output_path, 'w') as handle:
                    handle.write(svg_or_table)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import csv

def csv_string_to_data(csv_dump):
    data = list(csv.reader(csv_dump.splitlines()))
    return data[1:]


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    create_file('1234', 'a')
    create_file('1234', 'b')
    create_file('1234', 'stupid\'file,name')
    head, *data, footer, csv = run_rmlint('-S a', outputs=['csv'])

    (type_1, path_1, size_1, cksum_1),  \
    (type_2, path_2, size_2, cksum_2),  \
    (type_3, path_3, size_3, cksum_3) = \
        csv_string_to_data(csv)

    assert int(size_1) == 4
    assert int(size_2) == 4
    assert int(size_3) == 4

    assert path_1.endswith('/a')
    assert path_2.endswith('/b')
    assert path_3.endswith('/stupid\'file,name')

    assert type_1 == 'duplicate_file'
    assert type_2 == 'duplicate_file'
    assert type_3 == 'duplicate_file'

    assert cksum_1 != '0' * 32
    assert cksum_2 != '0' * 32
    assert cksum_3 != '0' * 32



#!/usr/bin/env python3
# encoding: utf-8

""" This file is part of rmlint.

rmlint is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

rmlint is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with rmlint.  If not, see <http://www.gnu.org/licenses/>.

Authors:

- Christopher <sahib> Pahl 2010-2017 (https://github.com/sahib)
- Daniel <SeeSpotRun> T.   2014-2017 (https://github.com/SeeSpotRun)
"""

# This is the python remover utility shipped inside the rmlint binary.
# The 200 lines source presented below is meant to be clean and hackable.
# It is intended to be used for corner cases where the built-in sh formatter
# is not enough or as an alternative to it. By default it works the same.
#
# Disable a few pylint warnings, in case someone integrates into scripts:
# pylint: disable=unused-argument,missing-docstring,invalid-name
# pylint: disable=redefined-outer-name,unused-variable

# Python2 compat:
from __future__ import print_function

import os
import sys
import pwd
import json
import shutil
import filecmp
import argparse
import subprocess

CURRENT_UID = os.geteuid()
CURRENT_GID = pwd.getpwuid(CURRENT_UID).pw_gid

USE_COLOR = sys.stdout.isatty() and sys.stderr.isatty()
COLORS = {
    'red':    "\x1b[0;31m" if USE_COLOR else "",
    'blue':   "\x1b[1;34m" if USE_COLOR else "",
    'green':  "\x1b[0;32m" if USE_COLOR else "",
    'yellow': "\x1b[0;33m" if USE_COLOR else "",
    'reset':  "\x1b[0m" if USE_COLOR else "",
}


def original_check(path, original, be_paranoid=True):
    try:
        stat_p, stat_o = os.stat(path), os.stat(original)
        if (stat_p.st_dev, stat_p.st_ino) == (stat_o.st_dev, stat_o.st_ino):
            print('{c[red]}Same inode; ignoring:{c[reset]} {o} <=> {p}'.format(
                c=COLORS, o=original, p=path))
            return False

        if stat_p.st_size != stat_o.st_size:
            print('{c[red]}Size differs; ignoring:{c[reset]} '
                  '{o} <=> {p}'.format(c=COLORS, o=original, p=path))
            return False

        if be_paranoid and not filecmp.cmp(path, original):
            print('{c[red]}Content differs; ignoring:{c[reset]} '
                  '{o} <=> {p}'.format(c=COLORS, o=original, p=path))
            return False

        return True
    except OSError as exc:
        print('{c[red]}{exc}{c[reset]}'.format(c=COLORS, exc=exc))
        return False


def handle_duplicate_dir(path, original, **kwargs):
    if not args.dry_run:
        shutil.rmtree(path)


def handle_duplicate_file(path, original, args, **kwargs):
    if original_check(path, original['path'], be_paranoid=args.paranoid):
        if not args.dry_run:
            os.remove(path)


def handle_unique_file(path, **kwargs):
    pass  # doesn't need any handling.


def handle_empty_dir(path, **kwargs):
    if not args.dry_run:
        os.rmdir(path)


def handle_empty_file(path, **kwargs):
    if not args.dry_run:
        os.remove(path)


def handle_nonstripped(path, **kwargs):
    if not args.dry_run:
        subprocess.call(["strip", "--strip-debug", path])


def handle_badlink(path, **kwargs):
    if not args.dry_run:
        os.remove(path)


def handle_baduid(path, **kwargs):
    if not args.dry_run:
        os.chown(path, args.user, -1)


def handle_badgid(path, **kwargs):
    if not args.dry_run:
        os.chown(path, -1, args.group)


def handle_badugid(path, **kwargs):
    if not args.dry_run:
        os.chown(path, args.user, args.group)


OPERATIONS = {
    "duplicate_dir": handle_duplicate_dir,
    "duplicate_file": handle_duplicate_file,
    "unique_file": handle_unique_file,
    "emptydir": handle_empty_dir,
    "emptyfile": handle_empty_file,
    "nonstripped": handle_nonstripped,
    "badlink": handle_badlink,
    "baduid": handle_baduid,
    "badgid": handle_badgid,
    "badugid": handle_badugid,
}


def exec_operation(item, original=None, args=None):
    try:
        OPERATIONS[item['type']](
            item['path'], original=original, item=item, args=args)
    except OSError as err:
        print('{c[red]}# {err}{c[reset]}'.format(
            err=err, c=COLORS
        ), file=sys.stderr)


MESSAGES = {
    'duplicate_dir':    '{c[yellow]}Deleting duplicate directory:',
    'duplicate_file':   '{c[yellow]}Deleting duplicate:',
    'unique_file':      'checking',
    'emptydir':         '{c[green]}Deleting empty directory:',
    'emptyfile':        '{c[green]}Deleting empty file:',
    'nonstripped':      '{c[green]}Stripping debug symbols:',
    'badlink':          '{c[green]}Deleting bad symlink:',
    'baduid':           '{c[green]}chown {u}',
    'badgid':           '{c[green]}chgrp {g}',
    'badugid':          '{c[green]}chown {u}:{g}',
}

ORIGINAL_MESSAGES = {
    'duplicate_file':   '{c[green]}Keeping original:  ',
    'duplicate_dir':    '{c[green]}Keeping original directory:  ',
}


def main(args, data):
    last_original_item = None

    # Process header and footer, if present
    header, footer = [], []
    if data[0].get('description'):
        header = data.pop(0)
    if data[-1].get('total_files'):
        footer = data.pop(-1)

    if not args.no_ask and not args.dry_run:
        print('rmlint was executed in the following way:\n',
              header.get('args'),
              '\n\nPress Enter to continue and perform modifications, '
              'or CTRL-C to exit.'
              '\nExecute this script with -d to disable this message.',
              file=sys.stderr)
        sys.stdin.read(1)

    for item in data:
        progress_prefix = '{c[blue]}[{p:3}%]{c[reset]} '.format(
            c=COLORS, p=item['progress'])

        if item['is_original']:
            msg = ORIGINAL_MESSAGES[item['type']].format(c=COLORS)
            print('{prog}{v}{c[reset]} {path}'.format(
                c=COLORS, prog=progress_prefix, v=msg, path=item['path']))
            last_original_item = item
            # Do not handle originals.
            continue

        msg = MESSAGES[item['type']].format(
            c=COLORS, u=args.user, g=args.group)
        print('{prog}{v}{c[reset]} {path}'.format(
            c=COLORS, prog=progress_prefix, v=msg, path=item['path']))
        exec_operation(item, original=last_original_item, args=args)

    print('{c[blue]}[100%] Done!{c[reset]}'.format(c=COLORS))


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='Handle the files in a JSON output of rmlint.'
    )

    parser.add_argument(
        'json_files', metavar='json_file', nargs='*', default=['.rmlint.json'],
        help='A JSON output of rmlint to handle (can be given multiple times)'
    )
    parser.add_argument(
        '-n', '--dry-run', action='store_true',
        help='Do not perform any modifications, just print what would be '
             'done. (implies -d)'
    )
    parser.add_argument(
        '-d', '--no-ask', action='store_true', default=False,
        help='Do not ask for confirmation before running.'
    )
    parser.add_argument(
        '-p', '--paranoid', action='store_true', default=False,
        help='Recheck that files are still identical before removing '
             'duplicates.'
    )
    parser.add_argument(
        '-u', '--user', type=int, default=CURRENT_UID,
        help='Numerical uid for chown operations'
    )
    parser.add_argument(
        '-g', '--group', type=int, default=CURRENT_GID,
        help='Numerical gid for chgrp operations'
    )

    args = parser.parse_args()
    json_docs = []
    for json_file in args.json_files:
        try:
            with open(json_file) as f:
                j = json.load(f)
            json_docs.append(j)
        except IOError as err:      # Cannot open file
            print(err, file=sys.stderr)
            sys.exit(-1)
        except ValueError as err:   # File is not valid JSON
            print('{}: {}'.format(err, json_file), file=sys.stderr)
            sys.exit(-1)

    try:
        if args.dry_run:
            print(
                '{c[green]}#{c[reset]} '
                'This is a dry run. Nothing will be modified.'.format(
                    c=COLORS
                )
            )

        for json_doc in json_docs:
            main(args, json_doc)

        if args.dry_run:
            print(
                '{c[green]}#{c[reset]} '
                'This was a dry run. Nothing was modified.'.format(
                    c=COLORS
                )
            )
    except KeyboardInterrupt:
        print('\ncanceled.')




#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *


@with_setup(usual_setup_func, usual_teardown_func)
def test_simple():
    full_path_a = create_file('x', '\t\r\"\b\f\\')
    full_path_b = create_file('x', '\"\t\n2134124')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 2

    for i in range(2):
        with open(data[i]['path']) as f:
            assert len(f.read()) == 1

    assert os.stat(full_path_a).st_size ==  data[0]['size']
    assert os.stat(full_path_b).st_size ==  data[1]['size']
    assert footer['total_lint_size'] == 1



#!/usr/bin/env python
# encoding: utf-8

"""
Chart rendering code and relevant Gtk widgets.
The chart is drawn via cairo and a tiny bit of math.
"""


# Stdlib:
import math
import colorsys

# Internal:
from shredder.util import size_to_human_readable
from shredder.tree import Column

# External:
import cairo

from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GLib

from gi.repository import Pango
from gi.repository import PangoCairo


ANGLE_LIMIT_TOOLTIP = math.pi / 32
ANGLE_LIMIT_VISIBLE = math.pi / 256


###########################################################
# NOTE: This code is inspired by the baobab code,         #
#       since it's very clean and a good read.            #
#       However this code is not ported, but rewritten.   #
###########################################################


def _draw_center_text(ctx, x, y, text, font_size=10, do_draw=True):
    '''Draw a text at the center of ctx/alloc.

    ctx: a cairo Context to draw to
    alloc: size of area.
    text: Actual text as unicode string
    font_size: Size of the font to take
    '''
    layout = PangoCairo.create_layout(ctx)
    font = Pango.FontDescription.from_string('Ubuntu Light')
    font.set_size(font_size * Pango.SCALE)
    layout.set_font_description(font)
    layout.set_markup(text, -1)
    layout.set_alignment(Pango.Alignment.CENTER)

    fw, fh = (num / Pango.SCALE / 2 for num in layout.get_size())
    if do_draw:
        ctx.move_to(x - fw, y - fh)
        PangoCairo.show_layout(ctx, layout)
        ctx.stroke()

    return fw, fh


def _draw_rounded(ctx, area, radius):
    """draws rectangles with rounded (circular arc) corners"""
    a, b, c, d = area
    mpi2 = math.pi / 2
    ctx.arc(a + radius, c + radius, radius, 2 * mpi2, 3 * mpi2)
    ctx.arc(b - radius, c + radius, radius, 3 * mpi2, 4 * mpi2)
    ctx.arc(b - radius, d - radius, radius, 0 * mpi2, 1 * mpi2)
    ctx.arc(a + radius, d - radius, radius, 1 * mpi2, 2 * mpi2)
    ctx.close_path()


TANGO_TABLE = [
    (0.000, 0.829, 0.94),
    (0.096, 0.747, 0.99),
    (0.104, 0.527, 0.91),
    (0.251, 0.775, 0.89),
    (0.590, 0.451, 0.82),
    (0.850, 0.279, 0.68)
]


def _hsv_by_degree(degree):
    """Convert degree (in rad) to a predefined color.
    Currently only one colorscheme is supported (Tango)
    """
    percent = max(0.0, min(1.0, degree / (2 * math.pi)))
    idx = percent * len(TANGO_TABLE)
    h, s, v = TANGO_TABLE[max(0, int(idx) - 1)]
    return h, s, v


def _draw_segment(
        ctx, alloc, layer, max_layers,
        deg_a, deg_b, is_selected, bg_col):
    """Draw a radial segment on the context ctx with the following params:

    layer: The segment layer to draw (or "how far from the midpoint we are")
    max_layers: How many layerst there are at max.
    deg_a: Starting angle of the segment in degree.
    deg_b: Ending angle of the segment in degree.

    The dimensions of ctx are given by size in pixels.
    """
    mid_x, mid_y = alloc.width / 2, alloc.height / 2
    mid = min(mid_x, mid_y)

    radius_a_px = (layer / (max_layers + 1)) * mid
    radius_b_px = radius_a_px + (1 / (max_layers + 1)) * mid

    if layer == 1:
        radius_a_px += (radius_b_px - radius_a_px) / 2

    # Draw the actual segment path
    ctx.arc(mid_x, mid_y, radius_a_px, deg_a, deg_b)
    ctx.arc_negative(mid_x, mid_y, radius_b_px, deg_b, deg_a)
    ctx.close_path()

    # Calculate the color as HSV
    h, s, v = _hsv_by_degree(deg_a / 2 + deg_b / 2)

    # "Fix" the color for some special cases
    s += 0.2 if is_selected else -0.07

    # Make more distant segments darker
    v *= 1.25 - (layer / max_layers / 1.5)

    if is_selected:
        v += 0.2

    # Fill it with the color: Add a bit of highlight on start & end
    # to round it up. This should stay rather subtle of course.
    pattern = cairo.RadialGradient(
        mid_x, mid_y, radius_a_px, mid_x, mid_y, radius_b_px
    )

    rn, gn, bn = colorsys.hsv_to_rgb(h, s, v)
    rh, gh, bh = colorsys.hsv_to_rgb(h, s + 0.05, v + 0.15)
    rl, gl, bl = colorsys.hsv_to_rgb(h, s - 0.05, v - 0.25)
    off = 42 / min(alloc.width, alloc.height)

    pattern.add_color_stop_rgb(0.0, rl, gl, bl)
    pattern.add_color_stop_rgb(off * 2, rn, gn, bn)
    pattern.add_color_stop_rgb(1 - off, rn, gn, bn)
    pattern.add_color_stop_rgb(1.0, rh, gh, bh)
    ctx.set_source(pattern)
    ctx.fill_preserve()

    # Draw a little same colored border around.
    thickness = min(alloc.width, alloc.height) / 160

    r, g, b = colorsys.hsv_to_rgb(h, s, v - 0.5)
    ctx.set_source_rgba(r, g, b, 0.5)
    ctx.set_line_width(thickness * 1.5)
    ctx.stroke_preserve()

    # Draw a (probably) white border around
    ctx.set_source_rgb(bg_col.red, bg_col.green, bg_col.blue)
    ctx.set_line_width(thickness)
    ctx.stroke()

    ctx.set_line_width(1)


def _draw_tooltip(ctx, alloc, x, y, dist, angle, text):
    """Draw a tooltip rooting at (x,y) and hitting the bounding box with `text`
    """
    # Draw the anchor circle on the segment
    ctx.set_source_rgba(0, 0, 0, 1.0)
    ctx.arc(x, y, 3, 0, 2 * math.pi)
    ctx.fill()

    # Guess the font width used for the tooltip text.
    fw, fh = _draw_center_text(ctx, 0, 0, text, do_draw=False)

    # Bounding box onto which the tooltips will be projected
    w2, h2 = alloc.width / 2 - dist - fw, alloc.height / 2 - dist - fh
    angle_norm = math.fmod(math.fabs(angle), math.pi / 2)

    # Check if we are in the second or last quadrant.
    # In that case we flip the dimensions in order
    # to make the flipping logic below work.
    if math.floor(angle / (math.pi / 2)) % 2:
        w2, h2 = h2, w2

    # Check where to paint the new point
    if angle_norm < math.atan2(h2, w2):
        new_x = w2
        new_y = math.tan(angle_norm) * new_x
    else:
        new_y = h2
        new_x = new_y / math.tan(angle_norm)

    # Flip the x/y into the right quadrant
    if angle < math.pi / 2:
        pass  # already in the right angle.
    elif angle < math.pi:
        new_x, new_y = -new_y, new_x
    elif angle < (math.pi + math.pi / 2):
        new_x, new_y = -new_x, -new_y
    elif angle < 2 * math.pi:
        new_x, new_y = new_y, -new_x

    # Finally, move from first quadrant to the whole.
    new_x += alloc.width / 2
    new_y += alloc.height / 2

    ctx.set_source_rgba(0, 0, 0, 0.3)
    ctx.move_to(x, y)
    ctx.line_to(new_x, new_y)
    ctx.stroke()

    ctx.set_source_rgba(0, 0, 0, 1.0)
    tip_w, tip_h = fw + 5, fh + 3
    _draw_rounded(
        ctx, (new_x - tip_w, new_x + tip_w, new_y - tip_h, new_y + tip_h), 3
    )
    ctx.fill_preserve()
    ctx.set_source_rgba(0.3, 0.3, 0.3, 1.0)
    ctx.stroke()

    ctx.set_source_rgba(0.8, 0.8, 0.8, 1.0)
    _draw_center_text(ctx, new_x, new_y, text, do_draw=True)


class Chart(Gtk.DrawingArea):
    """Base class for charts providing the basic interfaces and signals."""
    def __init__(self):
        Gtk.DrawingArea.__init__(self)
        self.connect('draw', self.on_draw)

        self.add_events(
            self.get_events() |
            Gdk.EventMask.POINTER_MOTION_MASK |
            Gdk.EventMask.BUTTON_PRESS_MASK
        )
        self.connect('motion-notify-event', self.on_motion)
        self.connect('button-press-event', self.on_button_press_event)

    ##############################
    # TO BE OVERWRITTEN BY CHILD #
    ##############################

    def on_draw(self, area, ctx):
        """Put your subclass draw code here."""
        pass

    def on_motion(self, area, event):
        """Executed on ever pointer motion."""
        pass

    def on_button_press_event(self, area, event):
        """Executed on every pointer or keyboard press."""
        pass


class Segment:
    """Helper and data class for a single segment in a RingChart."""
    def __init__(self, node, layer, degree, size, tooltip=None):
        self.node = node
        self.children = []
        self.layer, self.degree, self.size = layer, degree, size
        self.degree = math.fmod(self.degree, math.pi * 2)
        self.is_selected = False

        # Cut off too long tooltips:
        if tooltip is not None:
            tooltip = GLib.markup_escape_text(tooltip, -1)
            if len(tooltip) > 60:
                self.tooltip = tooltip[:60] + '...'
            else:
                self.tooltip = tooltip

    def draw(self, ctx, alloc, max_layers, bg_col):
        """Trigger the actual drawing of the segment."""
        _draw_segment(
            ctx, alloc,
            self.layer, max_layers,
            self.degree, self.degree + self.size,
            self.is_selected,
            bg_col
        )

    def hit(self, layer, deg):
        """Check if the segment was hit by a click,
        depending on a certain layer and angle.
        """
        if self.layer != layer:
            self.is_selected = False
        else:
            self.is_selected = self.degree <= deg <= self.degree + self.size

        return self.is_selected

    def middle_point(self, alloc, max_layers):
        """Calculate the middle point of the segment.

        The middle point is here defined as the mid between
        start and end degree with the radius between both ends.
        This is used to determine the place where to stick the
        tooltip.
        """
        # Middle point of the whole diagram
        mid_x, mid_y = alloc.width / 2, alloc.height / 2

        # Distance from (mid_x, mid_y) to middle point
        if self.layer == 1:
            offset = 0.75
        else:
            offset = 0.5

        rad = ((self.layer + offset) / (max_layers + 1)) * min(mid_x, mid_y)

        # Half of the degree range + start
        deg = self.degree + self.size / 2
        return mid_x + rad * math.cos(deg), mid_y + rad * math.sin(deg)

    def middle_angle(self):
        """Calculate an angle that goes through the mid of the segment."""
        return self.degree + self.size / 2


class RingChart(Chart):
    """Chart type for visualizing the node-tree as segmented ring.
    Each depth becomes one ring. Each segment has 0 to one parent.
    Size of the node determines the size of the segment.
    """
    def __init__(self):
        Chart.__init__(self)

        # Id of the tooltip timeout
        self._timeout_id = None
        self._segment_list = []
        self.max_layers = 0
        self.total_size = 1
        self._selected_segment = None
        self._last_root = None

    def recursive_angle(self, node, angle, offset, layer_offset=0):
        """Calculates the angles of the segments and stores them in a
        list that is ordered by Z-Depth, so the plot appears to be layered
        with the root circle on top. This resembles a depth first traversal.
        """
        self._segment_list.append(Segment(
            node, node.depth - layer_offset,
            offset, angle, node[Column.PATH]
        ))

        child_offset = offset
        for child in node.indices:
            node_size = node[Column.SIZE]

            if node_size != 0:
                child_angle = (child[Column.SIZE] / node_size) * angle
            else:
                child_angle = angle

            # Do not investigate smaller nodes:
            if child_angle > ANGLE_LIMIT_VISIBLE:
                self.recursive_angle(
                    child, child_angle, child_offset, layer_offset
                )
                # Remember deepest layer:
                self.max_layers = max(
                    child.depth - layer_offset + 1,
                    self.max_layers
                )

            child_offset += child_angle

    def find_root(self, node):
        """Iterate to the first child that has more than one children"""
        if len(node.children) > 1:
            return node

        for child in node.children.values():
            return self.find_root(child)

        # Default to the actual root:
        return node

    def render(self, root, overwrite_root=True):
        """Render `root` and all children of it as chart."""
        # Skip over duplicate full circles:
        virt_root = self.find_root(root)

        # Reset the segment list, fill it again by dfs and sort it.
        self._segment_list = []
        self.max_layers = 0
        self.recursive_angle(virt_root, 2 * math.pi, 0, virt_root.depth - 1)
        self._segment_list.sort(key=lambda node: node.layer)

        # Make sure we show the right total size
        self.total_size = virt_root[Column.SIZE]

        if overwrite_root:
            self._last_root = root

        # Make sure it gets rendered soon:
        self.queue_draw()

    def on_draw(self, area, ctx):
        """Actual signal callback that triggers all the drawing."""

        # May happen on empty charts:
        max_layers = self.max_layers
        if max_layers <= 0:
            max_layers = 1
            draw_empty = True
        else:
            draw_empty = False

        # Figure out the background color of the drawing area
        alloc = area.get_allocation()

        # Caluclate the font size of the inner label.
        # Make it smaller if not enough place but cut off at a size of 12
        inner_circle = (1.4 / max_layers)
        inner_circle *= min(alloc.width, alloc.height) / 2
        font_size = min(12, inner_circle / 3)

        # Draw the center text:
        _draw_center_text(
            ctx, alloc.width / 2, alloc.height / 2,
            '<span color="#333"><small>{size}</small></span>'.format(
                size=size_to_human_readable(self.total_size)
            ),
            font_size=font_size
        )

        bg = self.get_toplevel().get_style_context().get_background_color(0)

        if not draw_empty:
            for segment in reversed(self._segment_list):
                segment.draw(ctx, alloc, max_layers, bg)
        else:
            ctx.arc(
                alloc.width / 2, alloc.height / 2,
                inner_circle / 2, 0, 2 * math.pi
            )
            ctx.stroke()

        if self._selected_segment is None or draw_empty:
            return

        for segment in self._segment_list:
            if segment.layer != self._selected_segment.layer:
                continue

            if segment is not self._selected_segment:
                if segment.size < ANGLE_LIMIT_TOOLTIP:
                    continue

            x, y = segment.middle_point(alloc, max_layers)
            _draw_tooltip(
                ctx, alloc, x, y, 8,
                segment.middle_angle(),
                segment.tooltip
            )

    def on_tooltip_timeout(self, segment):
        """Called once the mouse stayed over a segment for a longer time.
        """
        if self._timeout_id:
            self._selected_segment = segment
        else:
            self._selected_segment = None

        self.queue_draw()
        self._timeout_id = None

    def _hit(self, area, event, click_only=False):
        """Check what segments were hitten by a GdkEvent"""
        alloc = area.get_allocation()
        mid_x, mid_y = alloc.width / 2, alloc.height / 2

        # Calculate the degree between the vectors
        # a = (event.x + m, event.y + m) and (0, 1)
        x, y = event.x - mid_x, event.y - mid_y
        xy_abs = math.sqrt(x * x + y * y)
        cos = x / xy_abs

        if y < 0:
            # upper half
            selected_deg = (2 * math.pi) - math.acos(cos)
        else:
            # lower half
            selected_deg = math.acos(cos)

        # Check which layer we are operating on.
        selected_layer = (self.max_layers + 1) / min(mid_x, mid_y)
        selected_layer = math.floor(xy_abs * selected_layer)
        if selected_layer == 0:
            return (True, None)

        hit_segment = None
        for segment in self._segment_list:
            if segment.hit(selected_layer, selected_deg):
                hit_segment = segment
                if click_only:
                    break

        return bool(hit_segment), hit_segment

    def on_motion(self, area, event):
        """Called on pointer motion."""
        hit, segment = self._hit(area, event)

        if self._timeout_id is not None:
            GLib.source_remove(self._timeout_id)
            self._timeout_id = None
            self._selected_segment = None

        if hit and segment:
            id_ = GLib.timeout_add(
                250, self.on_tooltip_timeout, segment
            )
            self._timeout_id = id_

        self.queue_draw()

    def on_button_press_event(self, area, event):
        """Called on pointer and keyboard events"""
        hit, segment = self._hit(area, event, click_only=True)
        if hit:
            if segment is not None:
                self.render(segment.node, overwrite_root=False)
            elif self._last_root:
                self.render(self._last_root, overwrite_root=False)


class ChartStack(Gtk.Stack):
    """Wrapper around the chart drawing area.
    Provides a loading screen and "nothing" found screen.
    Change between those are crossfaded.
    """
    LOADING = 'loading'
    CHART = 'chart'
    EMPTY = 'empty'

    def __init__(self):
        Gtk.Stack.__init__(self)
        self.set_transition_duration(750)
        self.set_transition_type(Gtk.StackTransitionType.CROSSFADE)

        # Make sure we don't stick on the border:
        self.set_border_width(5)

        self.spinner = Gtk.Spinner()
        self.spinner.start()
        self.add_named(self.spinner, ChartStack.LOADING)

        self.chart = RingChart()
        self.add_named(self.chart, ChartStack.CHART)

        self.empty_label = Gtk.Label(
            '<span font="90">✔</span>\nNothing found!'
        )
        self.empty_label.set_use_markup(True)
        self.empty_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )
        self.add_named(self.empty_label, ChartStack.EMPTY)

    def render(self, root):
        """Trigger all render procedure"""
        self.chart.render(root)


if __name__ == '__main__':
    def main():
        """Stupid test main"""
        from shredder.tree import PathTreeModel
        model = PathTreeModel(['/home/sahib'])

        def push(size, path):
            """Helper for pushing a dummy path"""
            model.add_path(path, Column.make_row({'size': size}), True)

        push(500, '/home/sahib/docs/stuff.pdf')

        for idx, size in enumerate((700, 600, 200)):
            push(size, '/home/sahib/docs/more/' + 'stuff.pdf-' + str(idx))

        for idx in range(50):
            push(10, '/home/sahib/docs/more/' + 'small.pdf-' + str(idx))

        for idx in range(10):
            push(100, '/home/sahib/' + 'dummy-' + str(idx))

        push(1000, '/home/sahib/music/1.mp3')
        push(1200, '/home/sahib/music/sub/2.mp3')
        push(1200, '/home/sahib/music/sub/3.mp3')
        push(600, '/home/sahib/music/sub/4.mp3')
        model.trie.sort(Column.SIZE)
        print(model.trie)

        area = RingChart()
        area.render(model.trie.root)

        win = Gtk.Window()
        win.set_size_request(300, 500)
        win.connect('destroy', Gtk.main_quit)
        win.add(area)
        win.show_all()

        Gtk.main()

    main()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

@with_setup(usual_setup_func, usual_teardown_func)
def test_negative_with_basename():
    create_file('xxx', 'a')
    create_file('xxx', 'b')
    head, *data, footer = run_rmlint('-b')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0
    assert footer['duplicates'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_positive_with_basename():
    create_file('xxx', 'a/test')
    create_file('xxx', 'b/test')
    head, *data, footer = run_rmlint('-b')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1


@with_setup(usual_setup_func, usual_teardown_func)
def test_negative_without_basename():
    create_file('xxx', 'a/test')
    create_file('xxx', 'b/test')
    head, *data, footer = run_rmlint('-B')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 0
    assert footer['duplicates'] == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_positive_without_basename():
    create_file('xxx', 'a/test1')
    create_file('xxx', 'b/test2')
    head, *data, footer = run_rmlint('-B')
    assert footer['total_files'] == 2
    assert footer['total_lint_size'] == 3
    assert footer['duplicates'] == 1



#!/usr/bin/env python3
# encoding: utf-8
from nose.plugins.attrib import attr
from nose import with_setup
from tests.utils import *

@attr('slow')
@with_setup(usual_setup_func, usual_teardown_func)
def test_manyfiles():

    # create heaps of identical files:
    numfiles = 1024 * 32 + 1
    for i in range(numfiles):
        create_file('xxx', 'file' + str(i).zfill(7))

    # create heaps of identical pairs:
    numpairs = 1024 * 32 + 1
    for i in range(numpairs):
        create_file(str(i), 'a' + str(i).zfill(7))
        create_file(str(i), 'b' + str(i).zfill(7))

    head, *data, footer = run_rmlint('')
    assert len(data) == numfiles + numpairs * 2



#!/usr/bin/env python
# encoding: utf-8

"""
This is a rather generic interface for building a settings view
from a GSettingsSchema, which is (usually) defined as XML File.

For shredder this file is `org.gnome.Shredder.gschema.xml` and contains a
description of all keys used by Shredder.  Keys are typed in GSettings, so
depending on the type a suitable widget is constructed.  Changes from outside
are displayed by the widget and changes of the widget cause a change of the
key's value.

For reference, keys can be changed on the commandline too:
$ gsettings --schemadir ~/.glib-schemas set org.gnome.Shredder traverse-depth 2
"""

# Stdlib:
import re
import logging

from operator import itemgetter
from functools import partial

# External:
from gi.repository import Gtk
from gi.repository import GLib

# Internal:
from shredder.util import View, SuggestedButton, DestructiveButton
from shredder.util import FileSizeRange, MultipleChoiceButton


LOGGER = logging.getLogger('settings')


################
# TYPE WIDGETS #
################


def boolean_widget(settings, key_name, *_):
    """Provide a simple widget for a boolean option key"""
    switch = Gtk.Switch()
    settings.bind(key_name, switch, 'active', 0)
    switch.set_active(settings.get_boolean(key_name))
    return switch


def numeric_widget(settings, key_name, *_, step=1):
    """Provide a single widget to change a numeric key"""
    # Use GSetting's "reflection" to get key metadata:
    schema = settings.get_property('settings-schema')
    key = schema.get_key(key_name)
    range_type, range_variant = key.get_range()

    range_wdgt = Gtk.SpinButton.new_with_range(0, 10 ** 10, step)

    # Configure the range of the values if given:
    if range_type == 'range':
        min_val, max_val = range_variant
        range_wdgt.set_range(min_val, max_val)

    settings.bind(key_name, range_wdgt, 'value', 0)
    func = settings.get_int if step == 1 else settings.get_double
    range_wdgt.set_value(func(key_name))
    return range_wdgt


def range_widget(settings, key_name, *_):
    """Create a range widget for key types like (tt)"""
    min_val, max_val = settings.get_value(key_name)
    widget = FileSizeRange(min_val, max_val)

    def setting_changed(*_):
        """GSettings -> Widget callback."""
        min_val, max_val = settings.get_value(key_name)
        widget.min_value = min_val
        widget.max_value = max_val

    def widget_changed(*_):
        """Widget -> GSettings callback."""
        min_val, max_val = widget.min_value, widget.max_value
        settings.set_value(key_name, GLib.Variant('(tt)', (min_val, max_val)))

    # Bind manually:
    settings.connect('changed::' + key_name, setting_changed)
    widget.connect('value-changed', widget_changed)

    return widget


def choice_widget(settings, key_name, summary, _):
    """Create a widget for choosing between a list of selections"""
    schema = settings.props.settings_schema
    key = schema.get_key(key_name)

    selected = settings.get_string(key_name)
    default = key.get_default_value().get_string()

    range_type, range_variant = key.get_range()
    if range_type != 'enum':
        LOGGER.warning('%s is not an enum.', key_name)
        return

    choices = list(range_variant)
    button = MultipleChoiceButton(choices, default, selected)
    button.connect(
        'row-selected',
        lambda _: settings.set_string(key_name, button.get_selected_choice())
    )

    settings.bind(key_name, button.value_label, 'choice', 0)
    return button


VARIANT_TO_WIDGET = {
    'b': boolean_widget,
    'i': partial(numeric_widget),
    'd': partial(numeric_widget, step=0.1),
    's': choice_widget,
    '(ii)': range_widget,
    '(tt)': range_widget
}


#####################
#    ACTUAL VIEW    #
#####################


class SettingsView(View):
    """Generic GSettingsView in a modern Gnome like appearance."""
    def __init__(self, app):
        View.__init__(
            self, app, sub_title='Configure how duplicates are searched'
        )

        self._grid = Gtk.Grid()
        self._grid.set_margin_start(30)
        self._grid.set_margin_end(40)
        self._grid.set_margin_top(5)
        self._grid.set_margin_bottom(15)
        self._grid.set_hexpand(True)
        self._grid.set_vexpand(True)
        self._grid.set_halign(Gtk.Align.FILL)
        self._grid.set_valign(Gtk.Align.FILL)
        self.add(self._grid)

        self.save_settings = False
        self.sections = {}
        self.metadata = {}

        self.appy_btn = SuggestedButton()
        self.deny_btn = DestructiveButton('Reset to defaults')

        self.appy_btn.connect('clicked', self.on_apply_settings)
        self.deny_btn.connect('clicked', self.on_reset_to_defaults)
        self.search_entry.connect(
            'search-changed', self.on_search_changed
        )

        # Initialize from current settings:
        self.build()

    def append_section(self, heading):
        """Append a new named section for multiple entries with `heading`"""
        box = Gtk.ListBox()
        box.set_selection_mode(Gtk.SelectionMode.NONE)
        box.set_size_request(350, -1)
        box.set_hexpand(True)

        frame = Gtk.Frame()
        frame.set_halign(Gtk.Align.FILL)
        frame.add(box)

        label = Gtk.Label()
        label.set_margin_top(30)
        label.set_markup(
            '<b>{}:</b>'.format(GLib.markup_escape_text(heading, -1))
        )
        label.set_halign(Gtk.Align.START)
        label.set_margin_bottom(2)

        self.sections[heading.lower()] = box
        self.metadata[heading.lower()] = {
            'label': label, 'frame': frame
        }
        self._grid.attach(label, 0, len(self._grid), 1, 1)
        self._grid.attach(frame, 0, len(self._grid), 1, 1)

    def append_entry(self, section, val_widget, key_name, summary, desc=None):
        """Append an entry to a named section.

        section: A previously inserted section name.
        val_widget: The widget to show the key in.
        summary: A short summary to show.
        desc: A longer description.
        """
        desc_label = Gtk.Label(desc or '')
        summ_label = Gtk.Label(summary or '')

        desc_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )

        for label in desc_label, summ_label:
            label.set_use_markup(True)
            label.set_halign(Gtk.Align.FILL)
            label.set_hexpand(True)
            label.set_halign(Gtk.Align.START)

        val_widget.set_halign(Gtk.Align.END)

        sub_grid = Gtk.Grid()
        sub_grid.attach(summ_label, 0, 0, 1, 1)
        sub_grid.attach(desc_label, 0, 1, 1, 1)
        sub_grid.attach(val_widget, 1, 0, 1, 2)
        sub_grid.set_border_width(3)

        listbox = self.sections[section.lower()]
        if len(listbox):
            sep_row = Gtk.ListBoxRow()

            # Needs gtk >= 3.14
            if hasattr(sep_row, 'set_activatable'):
                sep_row.set_activatable(False)

            sep_row.add(Gtk.Separator())
            listbox.insert(sep_row, -1)

        row = Gtk.ListBoxRow()
        row.add(sub_grid)
        row.set_can_focus(False)
        listbox.insert(row, -1)

        self.metadata[section.lower()][key_name] = {
            'summary': '' if summary is None else summary.lower(),
            'description': '' if desc is None else desc.lower(),
            'widget': row
        }

    def reset_to_defaults(self):
        """Reset whole view and keys to their defaults"""
        for key_name in self.app.settings.list_keys():
            self.app.settings.reset(key_name)

    def build(self):
        """Built all entries and sections"""
        gst = self.app.settings
        entry_rows = []

        for key_name in gst.list_keys():
            key = gst.get_property('settings-schema').get_key(key_name)
            variant_key = gst.get_value(key_name)

            # Try to find a way to render this option:
            constructor = VARIANT_TO_WIDGET.get(variant_key.get_type_string())
            if constructor is None:
                continue

            # Get the key summary and description:
            summary = '{}'.format(key.get_summary())

            # This is an extension of this code:
            if summary.startswith('[hidden]'):
                continue

            order, order_grep = 0, re.match(r'\[(\d+)]\s(.*)', summary)
            if order_grep is not None:
                order, summary = int(order_grep.group(1)), order_grep.group(2)

            description = key.get_description()
            if description:
                description = '<small>{desc}</small>'.format(
                    desc=key.get_description()
                )

            # Get a fitting, readily prepared configure widget
            val_widget = constructor(gst, key_name, summary, description)

            # Try to find the section name by the keyname.
            # This is an extension made by this code and not part of GSettings.
            section = ''
            if '-' in key_name:
                section, _ = key_name.split('-', maxsplit=1)

            entry_rows.append(
                (order, section, val_widget, key_name, summary, description)
            )

        for section in sorted(set([entry[1] for entry in entry_rows])):
            self.append_section(section.capitalize())

        for entry in sorted(entry_rows, key=itemgetter(0, 2)):
            self.append_entry(*entry[1:])

    ####################
    # SIGNAL CALLBACKS #
    ####################

    def on_search_changed(self, _):
        """Called once the user enteres a new search query."""
        query = self.search_entry.get_text().lower()

        def _set_vis(widget, state, lower):
            """Set opacity and sensitivity in one."""
            widget.set_sensitive(state)
            widget.set_opacity(1.0 if state else lower)

        for _, metadata in self.metadata.items():
            section_visible = 0

            for key_name, info in metadata.items():
                if key_name in ['label', 'frame']:
                    continue

                row = info['widget']
                if query in info['summary'] or query in info['description']:
                    section_visible += 1
                    _set_vis(row, True, 0.5)
                else:
                    _set_vis(row, False, 0.5)

            section_frame = metadata['frame']
            section_label = metadata['label']
            _set_vis(section_frame, section_visible > 0, 0.2)
            _set_vis(section_label, section_visible > 0, 0.2)

    def on_view_enter(self):
        """Called once the view is visible. Delay save of settings."""
        self.save_settings = False
        self.app.settings.delay()

        # Give the buttons a specific context meaning:
        self.on_key_changed(self.app.settings, None)
        self.app.settings.connect('changed', self.on_key_changed)

        self.add_header_widget(self.appy_btn)
        self.add_header_widget(self.deny_btn, Gtk.Align.START)

        # It's usually more useful when switching back to the latest view,
        # not to the view to the right since user might just have wanted
        # to have a quick settings change from anywhere in the application.
        self.app_window.views.switch_to_previous_next()

    def on_view_leave(self):
        """Called once the view gets out of sight. Revert or apply."""
        if self.save_settings:
            self.app.settings.apply()
        else:
            self.app.settings.revert()

        self.clear_header_widgets()

    def on_apply_settings(self, *_):
        """Callback for the apply button."""
        self.save_settings = True
        self.app_window.views.switch_to_previous()

    def on_reset_to_defaults(self, *_):
        """Callback for the reset button."""
        self.app.settings.revert()

        GLib.timeout_add(
            100, lambda: self.reset_to_defaults() or self.app.settings.delay()
        )

        self.save_settings = False
        self.app_window.views.switch_to_previous()

    def on_key_changed(self, settings, _):
        """Called when a key in GSettings changes."""
        is_sensitive = settings.get_has_unapplied()
        self.appy_btn.set_sensitive(is_sensitive)
        self.deny_btn.set_sensitive(is_sensitive)

    def on_default_action(self):
        """Called on Ctrl-Enter"""
        if self.appy_btn.is_sensitive():
            self.on_apply_settings()
        else:
            self.app_window.views.switch_to_previous()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

# These tests are only here to check if printing help works.
# Well, actually it's to increase coverage to be honest.


@with_setup(usual_setup_func, usual_teardown_func)
def test_help():
    yelp = subprocess.check_output(
        ['./rmlint', '--help'], stderr=subprocess.STDOUT
    ).decode('utf-8')
    assert 'man 1 rmlint' in yelp
    assert '--show-man' in yelp


@with_setup(usual_setup_func, usual_teardown_func)
def test_man():
    yelp = subprocess.check_output(
        ['./rmlint', '--show-man'], stderr=subprocess.STDOUT
    ).decode('utf-8')
    assert 'Pahl' in yelp
    assert 'Thomas' in yelp



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import stat

from itertools import chain, combinations


def create_file_with_perms(content, path, permissions):
    perms = 0
    if 'r' in permissions:
        perms |= stat.S_IRUSR

    if 'w' in permissions:
        perms |= stat.S_IWUSR

    if 'x' in permissions:
        perms |= stat.S_IXUSR

    create_file(content, path)
    os.chmod(os.path.join(TESTDIR_NAME, path), perms)


@with_setup(usual_setup_func, usual_teardown_func)
def test_combinations():
    # This test does not work when run as root.
    # root can read the files anyways.
    if runs_as_root():
        return

    # Generate all combinations of 'rwx' in a fun way:
    rwx = set(
        chain(
            "rwx",
            (''.join(s) for s in combinations('rwx', 2)),
            ["rwx"]
        )
    )

    create_file_with_perms('xxx', 'none', '')
    for perm in rwx:
        create_file_with_perms('xxx', perm, perm)

    files_created = len(rwx) + 1

    head, *data, footer = run_rmlint('')
    assert footer['duplicate_sets'] == 1
    assert footer['ignored_files'] == 0
    assert footer['total_files'] == files_created
    assert footer['duplicates'] == 3

    for perm_opt, dupes in zip('rwx', (3, 1, 1)):
        head, *data, footer = run_rmlint('--perms ' + perm_opt)
        assert footer['duplicate_sets'] == 1
        assert footer['ignored_files'] == 4
        assert footer['total_files'] == files_created - 4
        assert footer['duplicates'] == dupes

    head, *data, footer = run_rmlint('--perms rwx')
    assert footer['duplicate_sets'] == 0
    assert footer['ignored_files'] == 7
    assert footer['total_files'] == 1
    assert footer['duplicates'] == 0

    for perm_opt, dupes in zip(["rw", "rx", "wx"], (1, 1, 0)):
        head, *data, footer = run_rmlint('--perms ' + perm_opt)
        assert footer['duplicate_sets'] == dupes
        assert footer['ignored_files'] == 6
        assert footer['total_files'] == files_created - 6
        assert footer['duplicates'] == dupes



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *

import os
import re
import shutil
import tempfile


@with_setup(usual_setup_func, usual_teardown_func)
def test_backup():
    create_file('content', 'name_x')
    create_file('content', 'name_y')

    temp_dir = tempfile.mkdtemp()
    try:
        test_path = os.path.join(temp_dir, "xxx.json")

        for _ in range(10):
            _, *data, _ = run_rmlint("-o json:" + test_path, outputs=[])
            assert len(data) == 2

        for path in os.listdir(temp_dir):
            assert path.startswith("xxx.")

            if path == "xxx.json":
                continue

            assert re.match(
                r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(|\.\d+)Z',
                path[len("xxx."):-len(".json")]
            ) is not None
    finally:
        shutil.rmtree(temp_dir)



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
import os


def create_data(len, flips=None):
    data = ['0'] * len
    for flip in flips or []:
        data[flip] = '1'

    return ''.join(data)


@with_setup(usual_setup_func, usual_teardown_func)
def test_small_diffs():

    if use_valgrind():
        N = 32
    else:
        # Takes horribly long elsewhise
        N = 128

    create_file(create_data(len=N, flips=None), 'a')
    create_file(create_data(len=N, flips=[-1]), 'b')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 0

    create_file(create_data(len=N, flips=[+1]), 'a')
    create_file(create_data(len=N, flips=[-1]), 'b')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 0

    create_file(create_data(len=N, flips=[+1]), 'a')
    create_file(create_data(len=N, flips=[+1]), 'b')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 2

    for i in range(0, N // 2):
        create_file(create_data(len=N, flips=[+i]), 'a')
        create_file(create_data(len=N, flips=[-i]), 'b')
        head, *data, footer = run_rmlint('-S a')

        if i == N - i or i == 0:
            assert len(data) == 2
        else:
            assert len(data) == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_one_byte_file_negative():
    create_file('1', 'one')
    create_file('2', 'two')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 0


@with_setup(usual_setup_func, usual_teardown_func)
def test_one_byte_file_positive():
    create_file('1', 'one')
    create_file('1', 'two')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 2

@with_setup(usual_setup_func, usual_teardown_func)
def test_two_hardlinks():
    create_file('xxx', 'a')
    create_link('a', 'b')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 2
    assert footer['total_lint_size'] == 0

@with_setup(usual_setup_func, usual_teardown_func)
def test_two_external_hardlinks():
    create_file('xxx', 'a')
    create_file('xxx', 'b')
    create_dirs('sub')
    create_link('a', 'sub/a')
    create_link('a', 'sub/b')
    head, *data, footer = run_rmlint('-S a')

    assert len(data) == 4
    assert footer['total_lint_size'] == 3

    head, *data, footer = run_rmlint(
        "{t}/sub".format(t=TESTDIR_NAME),
        use_default_dir=False
    )

    # No effective lint: Removing any link will not save any disk space.
    assert len(data) == 2
    assert footer['total_lint_size'] == 0



#!/usr/bin/env python3
# encoding: utf-8
from nose.plugins.attrib import attr
from nose import with_setup
from tests.utils import *


# current shredder algorithm does not handle large size-groups at all
# well, due to pre-matching "optimisation"
# https://github.com/SeeSpotRun/rmlint/blob/448cb0c76cbb6178105556ede2bfd864c6f83af3/lib/checksum.c#L678-L730
# which degenerates into an inefficient O(n^2) lookup with large size groups
BLACKLIST = ['paranoid']

@attr('slow')
@with_setup(usual_setup_func, usual_teardown_func)
def test_collision_resistance():
    # test for at least 20 bits of collision resistancel
    # this should detect gross errors in checksum encoding...

    numfiles = 1024*1024
    for i in range(numfiles):
        create_file(i, str(i), write_binary=True)

    for algo in CKSUM_TYPES:
        if algo not in BLACKLIST:
            *_, footer = run_rmlint('--read-buffer-len=4 -a {}'.format(algo))
            assert footer['duplicates'] == 0, 'Unexpected hash collision for hash type {}'.format(algo)



#!/usr/bin/env python
# encoding: utf-8

"""
Editor view.

Provides means to view the generated script and edit it.
Also contains the dangerous run button that triggers the final run.
Once clicked, a counter will be shown, showing the total number
of killed files in terms of size.
"""


# Stdlib:
import os
import time
import logging

# External:
from gi.repository import Gtk
from gi.repository import GLib
from gi.repository import Pango
from gi.repository import GObject

try:
    from gi.repository import Polkit
except ImportError:
    Polkit = None

# Internal:
from shredder.util import View, IconButton, scrolled, size_to_human_readable
from shredder.util import MultipleChoiceButton, SuggestedButton
from shredder.runner import Script


LOGGER = logging.getLogger('editor')


REMOVED_LABEL = '''<big>{s}</big><small> {n} removed</small>
<small>Currently {t}</small> <b><big>{p}</big></b>
'''

#############################
# Try to load GtkSourceView #
#############################

try:
    from gi.repository import GtkSource

    def _create_source_view():
        """Create a suitable text view + buffer for showing a sh script."""
        LOGGER.info('Using GtkSourceView since we have it.')

        buffer_ = GtkSource.Buffer()
        buffer_.set_highlight_syntax(True)

        view = GtkSource.View()
        view.set_buffer(buffer_)
        view.set_show_line_numbers(True)
        view.set_show_line_marks(True)
        view.set_auto_indent(True)

        return view, buffer_

    class _SearchRun:
        """Represent a run through all matches relative a query."""
        def __init__(self, buffer_, query):
            settings = GtkSource.SearchSettings()
            settings.set_search_text(query)
            settings.set_case_sensitive(False)

            self.ctx = GtkSource.SearchContext.new(buffer_, settings)
            self.ctx.set_highlight(True)
            self.ctx.set_match_style(GtkSource.Style(underline=True))
            self._last_mark = None

        @property
        def query(self):
            """Query of this search run"""
            return self.ctx.get_settings().get_search_text()

        def next_hop(self, view):
            """Return a GtkTextMark pointing at the next find or None.

            All matching items will be highlighted by default.
            """
            # Find out at which position we were last:
            buffer_ = self.ctx.props.buffer
            if self._last_mark is None:
                offset = buffer_.get_start_iter()
            else:
                offset = buffer_.get_iter_at_mark(self._last_mark)

            # Trigger actual searching:
            self.ctx.forward_async(offset, None, self.on_forward_finish, view)

        def on_forward_finish(self, source, result, view):
            """Called once GtkSourceSearchContext has finished processing."""
            valid, start, end = self.ctx.forward_finish(result)
            self._last_mark = None

            # Select & remember mark for next hop
            if valid:
                buffer_ = self.ctx.props.buffer
                buffer_.select_range(start, end)
                self._last_mark = buffer_.create_mark(None, end, True)
                view.scroll_mark_onscreen(self._last_mark)

    def _set_source_style(view, style_name):
        """If supported, set a color scheme by name."""
        style = GtkSource.StyleSchemeManager.get_default().get_scheme(
            style_name
        )

        if style:
            buffer_ = view.get_buffer()
            buffer_.set_style_scheme(style)

    def _set_source_lang(view, lang):
        """If supported, set a syntax highlighter to use."""
        language = GtkSource.LanguageManager.get_default().get_language(lang)
        buffer_ = view.get_buffer()
        buffer_.set_language(language)

# Fallback to the normal Gtk.TextView if no GtkSource.View could be imported
# This is the bare minimum we support. It's neither pretty nor very useful.
except ImportError:

    def _create_source_view():
        """Create a suitable text view + buffer for showing a sh script."""
        LOGGER.info('No GtkSourceView found.')

        buffer_ = Gtk.TextBuffer()
        view = Gtk.TextView()
        return view, buffer_

    class _SearchRun:
        """Dummy search functor that does nothing."""
        def __init__(self, *_):
            self.query = None

        def next_hop(self, *_):
            """No-op"""
            pass

    def _set_source_style(*_):
        """If supported, set a color scheme by name."""
        pass  # Not supported.

    def _set_source_lang(*_):
        """If supported, set a syntax highlighter to use."""
        pass  # Not supported.


def _create_running_screen():
    """Helper to configure a spinner for the delete screen."""
    spinner = Gtk.Spinner()
    spinner.start()
    return spinner


def _create_finished_screen(callback):
    """Give the user a nice, warm feeling."""
    control_grid = Gtk.Grid()
    control_grid.set_hexpand(False)
    control_grid.set_vexpand(False)
    control_grid.set_halign(Gtk.Align.CENTER)
    control_grid.set_valign(Gtk.Align.CENTER)

    # Lies make the user feel comfortable:
    label = Gtk.Label(
        use_markup=True, label='''<span font="65">✔</span>

<big><b>All went well!</b></big>
''',
        justify=Gtk.Justification.CENTER
    )
    label.get_style_context().add_class('dim-label')

    go_back = IconButton('go-jump-symbolic', 'Go back to Script')
    go_back.set_halign(Gtk.Align.CENTER)
    go_back.connect(
        'clicked', lambda _: callback()
    )

    control_grid.attach(label, 0, 0, 1, 1)
    control_grid.attach_next_to(
        go_back, label, Gtk.PositionType.BOTTOM, 1, 1
    )

    return control_grid


class RunningLabel(Gtk.Label):
    """Centered large label showing a size sum and the current deleted path."""
    def __init__(self):
        Gtk.Label.__init__(self)

        # Basename is more important:
        self.set_ellipsize(Pango.EllipsizeMode.START)

        # Make it appeared a bit dimmed:
        self.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )
        self.set_justify(Gtk.Justification.CENTER)
        self._is_dry_run = False
        self.reset()

    def set_is_dry_run(self, is_it):
        """If it is a dry run, don't scare the user"""
        self._is_dry_run = is_it

    def push(self, prefix, path):
        """Push a new path to the label, removing the old one."""
        if prefix.lower() == 'keeping':
            return

        try:
            buf = os.stat(path)
            self._size_sum += buf.st_size
        except OSError:
            pass

        text = REMOVED_LABEL.format(
            t=prefix.lower(),
            s=size_to_human_readable(self._size_sum),
            p=GLib.markup_escape_text(path),
            n="<b>would be</b>" if self._is_dry_run else ""
        )
        self.set_markup(text)

    def reset(self):
        """Reset the counter to initial state (zero)"""
        self._size_sum = 0
        self.push('', '')


class RunButton(IconButton):
    """Customized run button that can change color."""
    dry_run = GObject.Property(type=bool, default=True)

    def __init__(self, icon, label, state_btn):
        IconButton.__init__(self, icon, label)
        self.state = state_btn
        self.state.connect('notify::active', self._toggle_dry_run)

        self.bind_property(
            'dry_run', self.state, 'active',
            GObject.BindingFlags.BIDIRECTIONAL |
            GObject.BindingFlags.SYNC_CREATE
        )

        self.state.set_active(True)
        self._toggle_dry_run(self.state)

    def set_sensitive(self, mode):
        btn_ctx = self.get_style_context()

        if mode:
            btn_ctx.add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
        else:
            btn_ctx.remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)

        IconButton.set_sensitive(self, mode)
        self.state.set_sensitive(mode)

    def _toggle_dry_run(self, *_):
        """Change the color and severeness of the button."""
        ctx = self.get_style_context()
        if not self.state.get_state():
            ctx.remove_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
            ctx.add_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION)
            self.set_markup("Go delete some files!")
        else:
            ctx.remove_class(Gtk.STYLE_CLASS_DESTRUCTIVE_ACTION)
            ctx.add_class(Gtk.STYLE_CLASS_SUGGESTED_ACTION)
            self.set_markup("Run in dry mode")


def _create_icon_stack():
    """Create a small widget that shows alternating icons."""
    icon_stack = Gtk.Stack()
    icon_stack.set_transition_type(
        Gtk.StackTransitionType.CROSSFADE
    )
    icon_stack.set_transition_duration(100)

    for name, symbol in (('warning', '⚠'), ('danger', '☠'), ('info', 'ℹ')):
        icon_label = Gtk.Label(
            use_markup=True,
            justify=Gtk.Justification.CENTER
        )
        icon_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )
        icon_label.set_markup(
            '<span font="65">{symbol}</span>'.format(symbol=symbol)
        )
        icon_stack.add_named(icon_label, name)

    return icon_stack


class ScriptSaverDialog(Gtk.FileChooserWidget):
    """GtkFileChooserWidget tailored for saving a `Script` instance."""

    __gsignals__ = {
        'saved': (GObject.SIGNAL_RUN_FIRST, None, ()),
    }

    def __init__(self, editor_view):
        Gtk.FileChooserWidget.__init__(self)

        self.editor_view = editor_view
        self.set_select_multiple(False)
        self.set_create_folders(False)
        self.set_action(Gtk.FileChooserAction.SAVE)
        self.set_do_overwrite_confirmation(True)

        self.file_type = MultipleChoiceButton(
            ['sh', 'json', 'csv'], 'sh', 'sh'
        )
        self.file_type.set_halign(Gtk.Align.START)
        self.file_type.set_hexpand(True)
        self.file_type.connect('row-selected', self.on_file_type_changed)
        self.file_type.props.margin_end = 10

        self.confirm = SuggestedButton('Save')
        self.confirm.connect('clicked', self.on_save_clicked)
        self.confirm.set_halign(Gtk.Align.END)
        self.confirm.set_hexpand(False)
        self.confirm.set_sensitive(False)
        self.confirm.props.margin_end = 10

        self.cancel_button = IconButton('window-close-symbolic', 'Cancel')
        self.cancel_button.connect('clicked', self.on_cancel_clicked)
        self.cancel_button.set_halign(Gtk.Align.END)
        self.cancel_button.set_hexpand(False)

        self.connect('selection-changed', self.on_selection_changed)

        file_type_label = Gtk.Label('<b>Filetype</b>')
        file_type_label.set_use_markup(True)
        file_type_label.props.margin_end = 5
        file_type_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )

        self.extra_box = Gtk.Grid()
        self.extra_box.attach(self.file_type, 0, 0, 1, 1)
        self.extra_box.attach(self.confirm, 1, 0, 1, 1)
        self.extra_box.attach_next_to(
            file_type_label,
            self.file_type,
            Gtk.PositionType.LEFT,
            1,
            1
        )

        self.extra_box.set_hexpand(True)
        self.extra_box.set_halign(Gtk.Align.FILL)

    def show_controls(self):
        """Show cancel, save and file type chooser buttons."""
        self.editor_view.add_header_widget(self.extra_box)
        self.editor_view.add_header_widget(
            self.cancel_button, align=Gtk.Align.START
        )

        self.update_file_suggestion()

    def update_file_suggestion(self):
        """Suggest a name for the script to save."""
        file_type = self.file_type.get_selected_choice() or 'sh'
        self.set_current_name(time.strftime('rmlint-%FT%T%z.' + file_type))

    def on_file_type_changed(self, _):
        """Called once the user chose a different format"""
        current_path = self.get_filename()
        if not current_path:
            self.update_file_suggestion()
        else:
            try:
                path, _ = current_path.rsplit('.', 1)
                self.set_current_name(
                    path + '.' + self.file_type.get_selected_choice() or ''
                )
            except ValueError:
                # No extension. Leave it.
                pass

    def _exit_from_save(self):
        """Preparation to go back to script view."""
        self.emit('saved')
        self.editor_view.clear_header_widgets()

    def on_cancel_clicked(self, _):
        """Signal handler for the cancel button."""
        self._exit_from_save()

    def on_save_clicked(self, _):
        """Called once the user clicked the `Save` button"""
        file_type = self.file_type.get_selected_choice()
        abs_path = self.get_filename()

        runner = self.editor_view.app_window.views['runner'].runner
        LOGGER.info('Saving script as `%s` to: %s', file_type, abs_path)
        runner.save(abs_path, file_type)
        self._exit_from_save()

    def on_selection_changed(self, _):
        """Called once a file or directory was clicked"""
        filename = self.get_filename()
        self.confirm.set_sensitive(bool(filename))

        # Make sure the user-typed extension gets set in teh type chooser also.
        name = self.get_current_name()
        *_, extension = name.rsplit('.', 1)
        self.file_type.set_selected_choice(extension)


class OverlaySaveButton(Gtk.Overlay):
    """Button box that contains two buttons in an overlay.
    The overlay is shown on top of the script editor.
    Buttons are: A unlock button for asking for root permissions
    and a save button to save the script somewhere.
    """

    __gsignals__ = {
        'save-clicked': (GObject.SIGNAL_RUN_FIRST, None, ()),
        'unlock-clicked': (GObject.SIGNAL_RUN_FIRST, None, ())
    }

    def __init__(self):
        Gtk.Overlay.__init__(self)

        self._box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        self._box.get_style_context().add_class(
            Gtk.STYLE_CLASS_LINKED
        )

        self._lock_button = Gtk.LockButton()
        self._lock_button.props.margin = 20
        self._lock_button.props.margin_end = 0
        self._lock_button.connect(
            'clicked', lambda _: self.emit('unlock-clicked')
        )

        if Polkit is not None:
            try:
                perm = Polkit.Permission.new_sync(
                    'org.freedesktop.accounts.user-administration',
                    Polkit.UnixProcess.new_for_owner(os.getpid(), 0, -1),
                    None
                )
                self._lock_button.set_permission(perm)
            except GLib.Error as err:
                LOGGER.warning('Unable to get polkit permissions: ' + str(err))


        self._save_button = IconButton(
            'folder-download-symbolic', 'Save to file'
        )
        self._save_button.props.margin = 20
        self._save_button.props.margin_start = 0
        self._save_button.connect(
            'clicked', lambda _: self.emit('save-clicked')
        )

        # Note: we're not showing the lock button yet,
        #       since we did not yet a policy for it.
        #
        # self._box.pack_start(self._lock_button, False, True, 0)
        self._box.pack_start(self._save_button, False, True, 0)
        self._box.set_hexpand(False)
        self._box.set_vexpand(False)
        self._box.set_halign(Gtk.Align.END)
        self._box.set_valign(Gtk.Align.END)
        self.add_overlay(self._box)


class EditorView(View):
    """Actual view class."""
    def __init__(self, win):
        View.__init__(self, win)

        self._last_runner = None
        self.script = Script.create_dummy()

        control_grid = Gtk.Grid()
        control_grid.set_hexpand(False)
        control_grid.set_vexpand(False)
        control_grid.set_halign(Gtk.Align.CENTER)
        control_grid.set_valign(Gtk.Align.CENTER)

        self.info_label = Gtk.Label(
            use_markup=True,
            justify=Gtk.Justification.CENTER
        )
        self.info_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )
        self.state_label = Gtk.Label(
            use_markup=True,
            justify=Gtk.Justification.CENTER
        )
        self.state_label.get_style_context().add_class(
            Gtk.STYLE_CLASS_DIM_LABEL
        )

        self.icon_stack = _create_icon_stack()

        self.text_view, buffer_ = _create_source_view()
        self.text_view.set_name('ShredderScriptEditor')
        self.text_view.set_vexpand(True)
        self.text_view.set_valign(Gtk.Align.FILL)
        self.text_view.set_hexpand(True)
        self.text_view.set_halign(Gtk.Align.FILL)
        self.save_button = OverlaySaveButton()
        self.save_button.add(scrolled(self.text_view))
        self.save_chooser = ScriptSaverDialog(self)

        def on_save_button_clicked(_):
            """Switch to the save dialog in the stack."""
            self.set_search_mode(False)
            self.left_stack.set_visible_child_name('chooser')
            self.save_chooser.show_controls()
            self.set_info_help_text()
            self.set_correct_icon()
            self.run_button.set_sensitive(False)

        def on_save_clicked(_):
            """Switch back when the user has saved."""
            self.left_stack.set_visible_child_name('script')
            self.set_info_review_text()
            self.set_correct_icon()
            self.run_button.set_sensitive(True)

        self.save_button.connect(
            'save-clicked', on_save_button_clicked
        )

        self.save_chooser.connect(
            'saved', on_save_clicked
        )

        buffer_.create_tag("original", weight=Pango.Weight.BOLD)
        buffer_.create_tag("normal")

        self.run_label = RunningLabel()
        self.run_label.set_hexpand(False)
        self.run_label.set_halign(Gtk.Align.FILL)

        self.left_stack = Gtk.Stack()
        self.left_stack.set_transition_type(
            Gtk.StackTransitionType.SLIDE_UP
        )

        spinner = Gtk.Spinner()
        spinner.start()

        self.left_stack.add_named(spinner, 'loading')
        self.left_stack.add_named(self.save_button, 'script')
        self.left_stack.add_named(self.save_chooser, 'chooser')
        self.left_stack.add_named(scrolled(self.run_label), 'list')

        separator = Gtk.Separator(orientation=Gtk.Orientation.VERTICAL)
        left_pane = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
        left_pane.pack_start(self.left_stack, True, True, 0)
        left_pane.pack_start(separator, False, False, 0)

        self.state_btn = Gtk.Switch()
        self.state_btn.set_hexpand(False)
        self.state_btn.set_halign(Gtk.Align.CENTER)
        self.run_button = RunButton(
            'user-trash-symbolic', 'Run Script!', self.state_btn
        )

        self.run_button.connect('clicked', self.on_run_script_clicked)
        self.run_button.set_halign(Gtk.Align.CENTER)
        self.set_info_review_text()

        def on_dry_run_changed(*_):
            self.set_correct_icon()
            self.set_info_review_text()

        self.run_button.connect('notify::dry-run', on_dry_run_changed)

        control_grid.attach(self.info_label, 0, 0, 1, 1)
        control_grid.attach_next_to(
            self.run_button, self.info_label, Gtk.PositionType.BOTTOM, 1, 1
        )
        control_grid.attach_next_to(
            self.icon_stack, self.info_label, Gtk.PositionType.TOP, 1, 1
        )
        control_grid.attach_next_to(
            self.state_label, self.run_button, Gtk.PositionType.BOTTOM, 1, 1
        )
        control_grid.attach_next_to(
            self.state_btn, self.state_label, Gtk.PositionType.BOTTOM, 1, 1
        )
        control_grid.set_border_width(15)

        self.stack = Gtk.Stack()
        self.stack.set_transition_type(Gtk.StackTransitionType.SLIDE_UP)

        self.stack.add_named(control_grid, 'danger')
        self.stack.add_named(_create_running_screen(), 'progressing')

        self.stack.add_named(
            _create_finished_screen(self._switch_back), 'finished'
        )

        self.left_stack.set_visible_child_name('script')

        paned = Gtk.Paned()
        paned.pack1(left_pane, True, True)
        paned.pack2(self.stack, True, True)
        paned.props.position = 920
        self.add(paned)

        self._last_search = None
        self.search_entry.connect(
            'search-changed', self.on_search_changed
        )

        try:
            self.search_entry.connect(
                'next-match', self.on_search_changed
            )
        except TypeError:
            LOGGER.warning('Old gtk version; skipping through matches will not work.')

    def set_correct_icon(self):
        """Set the correct icon of icon_stack (either warning, skull or info)"""
        icon_name = 'info'

        if self.left_stack.get_visible_child_name() == 'script':
            if self.run_button.dry_run:
                icon_name = 'warning'
            else:
                icon_name = 'danger'

        self.icon_stack.set_visible_child_name(icon_name)

    def set_info_review_text(self):
        """Set the normal 'Review the script' text."""
        self.state_btn.show_all()
        self.info_label.set_markup('''


<big><b>Review the script on the left!</b></big>
When done, click the `Run Script` button below.
\n\n''')

        if self.run_button.dry_run:
            self.state_label.set_markup('''

<small><i>
I will currently only do a </i><b>dry run</b>.<i>
Please toggle the switch below to actually delete files:</i></small>
            ''')
        else:
            self.state_label.set_markup('''

<small><i>
Will run in </i><b>destructive mode and will delete files!</b><i>
Please toggle the switch below to go back to dry run:</i></small>
            ''')

    def set_info_help_text(self):
        """Be a bit more helpful on the help dialog."""
        self.info_label.set_markup('''
<big><b>Save the script for later!</b></big>

It can be executed via <span font_family="monospace">./rmlint.sh</span>
Or you can replay the output later with:
    <span font_family="monospace">rmlint --replay /path/to/file.json</span>
''')
        self.state_label.set_markup('''


<small><b>Note:</b> You can select the filetype in the header.</small>

''')
        self.state_btn.hide()


    def _switch_back(self):
        """Switch back from delete-view to script view"""
        self.switch_to_script()

    def switch_to_script(self):
        """Read and show the script."""
        self.sub_title = 'Check the results'
        GLib.idle_add(
            lambda: self.left_stack.set_visible_child_name('script')
        )
        buffer_ = self.text_view.get_buffer()
        buffer_.set_text(self.script.read())

        # Make sure it gets colored again:
        _set_source_style(self.text_view, 'solarized-light')
        _set_source_lang(self.text_view, 'sh')
        self.stack.set_visible_child_name('danger')
        self.stack.set_sensitive(True)

    def on_search_changed(self, _):
        """Called once the user enteres a new search query."""
        query = self.search_entry.get_text().lower()
        buffer_ = self.text_view.get_buffer()

        # If query is empty, just deselect everything.
        if not query:
            buffer_.select_range(
                buffer_.get_start_iter(),
                buffer_.get_start_iter()
            )

        # Check if query changed and if we need to get a new search ctx
        if self._last_search is None or self._last_search.query != query:
            self._last_search = _SearchRun(buffer_, query)

        # Jump one position ahead (or do initial search)
        self._last_search.next_hop(self.text_view)

    def on_run_script_clicked(self, _):
        """The critical function callback that is run when action is done."""
        self.set_search_mode(False)
        self.sub_title = 'Shreddering. Cross fingers!'
        self.stack.set_visible_child_name('progressing')
        self.left_stack.set_visible_child_name('list')

        LOGGER.info('Running script.')
        self.run_label.reset()
        self.run_label.set_is_dry_run(self.run_button.dry_run)
        self.script.run(dry_run=self.run_button.dry_run)

    def on_view_enter(self):
        """Called once the view becomes visible."""
        self.run_button.set_sensitive(True)

        # Re-read the script.
        runner = self.app_window.views['runner'].runner
        if runner is not None and self._last_runner is not runner:
            runner.connect('replay-finished', self.on_replay_finish, runner)
            self._last_runner = runner
            self.left_stack.set_visible_child_name('loading')
            self.stack.set_sensitive(False)

    def on_replay_finish(self, _, runner):
        """Called once ``rmlint --replay`` finished running."""
        LOGGER.info('Loading script from temporary directory')
        self.override_script(Script(runner.get_sh_path()))

    def on_default_action(self):
        """Called on Ctrl-Enter"""
        visible_screen = self.stack.get_visible_child_name()
        if visible_screen == 'danger':
            self.on_run_script_clicked(None)
        elif visible_screen == 'finished':
            self.switch_to_script()

    def override_script(self, script):
        """This method is for testing and cmdline use only."""
        LOGGER.info('Loading developer-defined script.')
        self.script = script

        self.script.connect(
            'line-read',
            lambda _, prefix, line: self.run_label.push(prefix, line)
        )

        self.script.connect(
            'script-finished',
            lambda *_: self.stack.set_visible_child_name('finished')
        )
        self.switch_to_script()



#!/usr/bin/env python3
# encoding: utf-8
from nose import with_setup
from tests.utils import *
from subprocess import STDOUT, check_output
from parameterized import parameterized

import json
import os


@with_setup(usual_setup_func, usual_teardown_func)
def test_stdin_read():
    path_a = create_file('1234', 'a') + '\n'
    path_b = create_file('1234', 'b') + '\n'
    path_c = create_file('1234', '.hidden') + '\n'

    subdir = 'look-in-here'
    create_file('1234', subdir + '/c')
    subdir_path = os.path.join(TESTDIR_NAME, subdir)

    proc = subprocess.Popen(
        ['./rmlint', '-', subdir_path, '-o', 'json', '-S', 'a', '--hidden'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE
    )
    data, _ = proc.communicate((path_a + path_b + path_c).encode('utf-8'))
    head, *data, footer = json.loads(data.decode('utf-8'))

    assert data[0]['path'].endswith('.hidden')
    assert data[1]['path'].endswith('a')
    assert data[2]['path'].endswith('b')
    assert data[3]['path'].endswith('c')
    assert footer['total_lint_size'] == 12

@with_setup(usual_setup_func, usual_teardown_func)
def test_stdin_read_newlines():
    path_a = create_file('1234', 'a') + '\0'
    path_b = create_file('1234', 'name\nwith\nnewlines') + '\0'
    path_c = create_file('1234', '.hidden') + '\0'

    subdir = 'look-in-here'
    create_file('1234', subdir + '/c')
    subdir_path = os.path.join(TESTDIR_NAME, subdir)

    proc = subprocess.Popen(
        ['./rmlint', '-0', subdir_path, '-o', 'json', '-S', 'a', '--hidden'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE
    )
    data, _ = proc.communicate((path_a + path_b + path_c).encode('utf-8'))
    head, *data, footer = json.loads(data.decode('utf-8'))

    assert data[0]['path'].endswith('.hidden')
    assert data[1]['path'].endswith('a')
    assert data[2]['path'].endswith('c')
    assert data[3]['path'].endswith('newlines')
    assert footer['total_lint_size'] == 12

@with_setup(usual_setup_func, usual_teardown_func)
def test_path_starting_with_dash():
    subdir = '-look-in-here'
    create_file('1234', subdir + '/a')
    create_file('1234', subdir + '/b')

    cwd = os.getcwd()

    try:
        os.chdir(TESTDIR_NAME)
        data = check_output(
            [cwd + '/rmlint', '-o', 'json', '-S', 'a', '--', subdir],
            stderr=STDOUT
        )
    finally:
        os.chdir(cwd)

    head, *data, footer = json.loads(data.decode('utf-8'))

    assert data[0]['path'].endswith('a')
    assert data[1]['path'].endswith('b')
    assert footer['total_lint_size'] == 4


# Regression test for https://github.com/sahib/rmlint/issues/400
# Do not search in current directory when piped empty input.
@parameterized([("-", ), ("-0", )])
@with_setup(usual_setup_func, usual_teardown_func)
def test_stdin_empty(stdin_opt):
    create_file('1234', 'a')
    create_file('1234', 'b')

    cwd = os.getcwd()

    try:
        os.chdir(TESTDIR_NAME)
        proc = subprocess.Popen(
            [cwd + '/rmlint', stdin_opt, '-o', 'json'],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE
        )
        data, _ = proc.communicate("")
        head, *data, footer = json.loads(data.decode('utf-8'))
    finally:
        os.chdir(cwd)

    assert len(data) == 0
